from typing import Literal, Type

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.enum import SamplingMethodEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    CategoricalMolecularInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
)
from bofire.data_models.strategies.samplers.sampler import SamplerStrategy


class PolytopeSampler(SamplerStrategy):
    """Sampler that generates samples from a Polytope defined by linear equality and ineqality constraints.

    Attributes:
        domain (Domain): Domain defining the constrained input space
        fallback_sampling_method: SamplingMethodEnum, optional): Method to use for sampling when no
            constraints are present. Defaults to UNIFORM.
    """

    type: Literal["PolytopeSampler"] = "PolytopeSampler"
    fallback_sampling_method: SamplingMethodEnum = SamplingMethodEnum.UNIFORM

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            LinearInequalityConstraint,
            LinearEqualityConstraint,
            NChooseKConstraint,
        ]

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            ContinuousInput,
            ContinuousOutput,
            CategoricalInput,
            DiscreteInput,
            CategoricalDescriptorInput,
            CategoricalMolecularInput,
        ]



import pandas as pd
from pydantic.types import PositiveInt

import bofire.data_models.strategies.api as data_models
from bofire.data_models.features.api import CategoricalInput, Input
from bofire.strategies.doe.design import (
    find_local_max_ipopt,
    find_local_max_ipopt_BaB,
    find_local_max_ipopt_exhaustive,
)
from bofire.strategies.doe.utils_categorical_discrete import (
    design_from_new_to_original_domain,
    discrete_to_relaxable_domain_mapper,
    nchoosek_to_relaxable_domain_mapper,
)
from bofire.strategies.strategy import Strategy


class DoEStrategy(Strategy):
    """Strategy for design of experiments. This strategy is used to generate a set of
    experiments for a given domain.
    The experiments are generated via minimization of the D-optimality criterion.

    """

    def __init__(
        self,
        data_model: data_models.DoEStrategy,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.formula = data_model.formula
        self.data_model = data_model
        self._partially_fixed_candidates = None
        self._fixed_candidates = None

    def set_candidates(self, candidates: pd.DataFrame):
        original_columns = self.domain.get_feature_keys(includes=Input)
        to_many_columns = []
        for col in candidates.columns:
            if col not in original_columns:
                to_many_columns.append(col)
        if len(to_many_columns) > 0:
            raise AttributeError(
                f"provided candidates have columns: {*to_many_columns,},  which do not exist in original domain"
            )

        to_few_columns = []
        for col in original_columns:
            if col not in candidates.columns:
                to_few_columns.append(col)
        if len(to_few_columns) > 0:
            raise AttributeError(
                f"provided candidates are missing columns: {*to_few_columns,} which exist in original domain"
            )

        self._candidates = candidates

    def _ask(self, candidate_count: PositiveInt) -> pd.DataFrame:
        all_new_categories = []

        # map categorical/ discrete Domain to a relaxable Domain
        new_domain, new_categories, new_discretes = discrete_to_relaxable_domain_mapper(
            self.domain
        )
        all_new_categories.extend(new_categories)

        # check for NchooseK constraint and solve the problem differently depending on the strategy
        if self.data_model.optimization_strategy != "partially-random":
            (
                new_domain,
                new_categories,
            ) = nchoosek_to_relaxable_domain_mapper(new_domain)
            all_new_categories.extend(new_categories)

        # here we adapt the (partially) fixed experiments to the new domain
        fixed_experiments_count = 0
        _candidate_count = candidate_count
        adapted_partially_fixed_candidates = self._transform_candidates_to_new_domain(
            new_domain, self.candidates
        )

        if self.candidates is not None:
            fixed_experiments_count = self.candidates.notnull().all(axis=1).sum()
            _candidate_count = candidate_count + fixed_experiments_count

        num_binary_vars = len([var for group in new_categories for var in group])
        num_discrete_vars = len(new_discretes)

        if (
            self.data_model.optimization_strategy == "relaxed"
            or (num_binary_vars == 0 and num_discrete_vars == 0)
            or (
                self.data_model.optimization_strategy == "partially-random"
                and num_binary_vars == 0
                and num_discrete_vars == 0
            )
        ):
            design = find_local_max_ipopt(
                new_domain,
                self.formula,
                n_experiments=_candidate_count,
                fixed_experiments=None,
                partially_fixed_experiments=adapted_partially_fixed_candidates,
            )
        # todo adapt to when exhaustive search accepts discrete variables
        elif (
            self.data_model.optimization_strategy == "exhaustive"
            and num_discrete_vars == 0
        ):
            design = find_local_max_ipopt_exhaustive(
                domain=new_domain,
                model_type=self.formula,
                n_experiments=_candidate_count,
                fixed_experiments=None,
                verbose=self.data_model.verbose,
                partially_fixed_experiments=adapted_partially_fixed_candidates,
                categorical_groups=all_new_categories,
                discrete_variables=new_discretes,
            )
        elif self.data_model.optimization_strategy in [
            "branch-and-bound",
            "default",
            "partially-random",
        ]:
            design = find_local_max_ipopt_BaB(
                domain=new_domain,
                model_type=self.formula,
                n_experiments=_candidate_count,
                fixed_experiments=None,
                verbose=self.data_model.verbose,
                partially_fixed_experiments=adapted_partially_fixed_candidates,
                categorical_groups=all_new_categories,
                discrete_variables=new_discretes,
            )
        elif self.data_model.optimization_strategy == "iterative":
            # a dynamic programming approach to shrink the optimization space by optimizing one experiment at a time
            assert (
                _candidate_count is not None
            ), "strategy iterative requires number of experiments to be set!"

            num_adapted_partially_fixed_candidates = 0
            if adapted_partially_fixed_candidates is not None:
                num_adapted_partially_fixed_candidates = len(
                    adapted_partially_fixed_candidates
                )
            design = None
            for i in range(_candidate_count):
                design = find_local_max_ipopt_BaB(
                    domain=new_domain,
                    model_type=self.formula,
                    n_experiments=num_adapted_partially_fixed_candidates + i + 1,
                    fixed_experiments=None,
                    verbose=self.data_model.verbose,
                    partially_fixed_experiments=adapted_partially_fixed_candidates,
                    categorical_groups=all_new_categories,
                    discrete_variables=new_discretes,
                )
                adapted_partially_fixed_candidates = pd.concat(
                    [
                        adapted_partially_fixed_candidates,
                        design.round(6).tail(1),
                    ],
                    axis=0,
                    ignore_index=True,
                )
                print(
                    f"Status: {i+1} of {_candidate_count} experiments determined \n"
                    f"Current experimental plan:\n {design_from_new_to_original_domain(self.domain, design)}"
                )

        else:
            raise RuntimeError("Could not find suitable optimization strategy")

        # mapping the solution to the variables from the original domain
        transformed_design = design_from_new_to_original_domain(self.domain, design)

        return transformed_design.iloc[fixed_experiments_count:, :].reset_index(drop=True)  # type: ignore

    def has_sufficient_experiments(
        self,
    ) -> bool:
        """Abstract method to check if sufficient experiments are available.

        Returns:
            bool: True if number of passed experiments is sufficient, False otherwise
        """
        return True

    def _transform_candidates_to_new_domain(self, new_domain, candidates):
        if candidates is not None:
            intermediate_candidates = candidates.copy()
            missing_columns = [
                key
                for key in new_domain.inputs.get_keys()
                if key not in candidates.columns
            ]

            for col in missing_columns:
                intermediate_candidates.insert(0, col, None)

            cat_columns = self.domain.get_features(includes=CategoricalInput)
            for cat in cat_columns:
                for row_index, c in enumerate(intermediate_candidates[cat.key].values):
                    if pd.isnull(c):
                        continue
                    if c not in cat.categories:  # type: ignore
                        raise AttributeError(
                            f"provided value {c} for categorical variable {cat.key} does not exist in the corresponding categories {cat.categories}"  # type: ignore
                        )
                    intermediate_candidates.loc[row_index, cat.categories] = 0  # type: ignore
                    intermediate_candidates.loc[row_index, c] = 1

            intermediate_candidates = intermediate_candidates.drop(
                [cat.key for cat in cat_columns], axis=1
            )

            adapted_partially_fixed_candidates = pd.concat(
                [
                    intermediate_candidates[candidates.notnull().all(axis=1)],
                    intermediate_candidates[candidates.isnull().any(axis=1)],
                ]
            )
            return adapted_partially_fixed_candidates
        return None



import inspect
from typing import Dict, List, Literal, Optional, Tuple, Type

import numpy as np
import pandas as pd
from pydantic.types import NonNegativeInt

import bofire.data_models.strategies.api as data_models
from bofire.data_models.constraints.api import (
    Constraint,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
)
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput, Feature
from bofire.data_models.objectives.api import (
    MaximizeObjective,
    MinimizeObjective,
    Objective,
)
from bofire.strategies.api import BotorchStrategy, PredictiveStrategy, Strategy


class DummyStrategyDataModel(data_models.BotorchStrategy):
    type: Literal["DummyStrategyDataModel"] = "DummyStrategyDataModel"

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        return my_type in [
            LinearEqualityConstraint,
            LinearInequalityConstraint,
        ]

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [ContinuousInput, ContinuousOutput]

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        return my_type in [
            MaximizeObjective,
            MinimizeObjective,
        ]


class DummyStrategy(Strategy):
    def __init__(
        self,
        data_model: DummyStrategyDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _init_domain(
        self,
    ) -> None:
        pass

    def _tell(
        self,
    ) -> None:
        pass

    def _ask(
        self,
        candidate_count: int,
    ) -> Tuple[pd.DataFrame, List[dict]]:
        raise NotImplementedError(
            f"{inspect.stack()[0][3]} not implemented for {self.__class__.__name__}"
        )

    def _choose_from_pool(
        self,
        candidate_pool: pd.DataFrame,
        candidate_count: Optional[NonNegativeInt] = None,
    ) -> pd.DataFrame:
        candidates = candidate_pool.sample(candidate_count, replace=False)
        for feat in self.domain.outputs.get_by_objective(Objective):
            candidates[f"{feat.key}_pred"] = np.nan
            candidates[f"{feat.key}_sd"] = np.nan
            candidates[f"{feat.key}_des"] = np.nan
        return candidates

    def has_sufficient_experiments(
        self,
    ) -> bool:
        return len(self.experiments) >= 3


class DummyPredictiveStrategyDataModel(data_models.Strategy):
    type: Literal[
        "DummyPredictiveStrategyDataModel"
    ] = "DummyPredictiveStrategyDataModel"

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        return my_type in [
            LinearEqualityConstraint,
            LinearInequalityConstraint,
        ]

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [ContinuousInput, ContinuousOutput]

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        return my_type in [
            MaximizeObjective,
            MinimizeObjective,
        ]


class DummyPredictiveStrategy(PredictiveStrategy):
    def __init__(
        self,
        data_model: DummyPredictiveStrategyDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _init_domain(
        self,
    ) -> None:
        pass

    def _tell(
        self,
    ) -> None:
        pass

    def _fit(self, transformed: pd.DataFrame):
        pass

    def _predict(self, experiments: pd.DataFrame):
        return (
            np.ones([len(experiments), len(self.domain.outputs)]) * 4,
            np.ones([len(experiments), len(self.domain.outputs)]) * 5,
        )

    def _ask(
        self,
        candidate_count: int,
    ) -> Tuple[pd.DataFrame, List[dict]]:
        raise NotImplementedError(
            f"{inspect.stack()[0][3]} not implemented for {self.__class__.__name__}"
        )

    def _choose_from_pool(
        self,
        candidate_pool: pd.DataFrame,
        candidate_count: Optional[NonNegativeInt] = None,
    ) -> pd.DataFrame:
        candidates = candidate_pool.sample(candidate_count, replace=False)
        for feat in self.domain.outputs.get_by_objective(Objective):
            candidates[f"{feat.key}_pred"] = np.nan
            candidates[f"{feat.key}_sd"] = np.nan
            candidates[f"{feat.key}_des"] = np.nan
        return candidates

    @property
    def input_preprocessing_specs(self):
        return {}

    def has_sufficient_experiments(
        self,
    ) -> bool:
        return len(self.experiments) >= 3


class DummyBotorchPredictiveStrategy(BotorchStrategy):
    def __init__(
        self,
        data_model: DummyStrategyDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _init_domain(
        self,
    ) -> None:
        pass

    def _predict(self, experiments: pd.DataFrame):
        return (
            np.ones([len(experiments), len(self.domain.outputs)]) * 4,
            np.ones([len(experiments), len(self.domain.outputs)]) * 5,
        )

    def _ask(
        self,
        candidate_count: int,
    ) -> Tuple[pd.DataFrame, List[dict]]:
        raise NotImplementedError(
            f"{inspect.stack()[0][3]} not implemented for {self.__class__.__name__}"
        )

    def _get_acqfs(self, n: int):
        pass

    def _choose_from_pool(
        self,
        candidate_pool: pd.DataFrame,
        candidate_count: Optional[NonNegativeInt] = None,
    ) -> pd.DataFrame:
        candidates = candidate_pool.sample(candidate_count, replace=False)
        for feat in self.domain.outputs.get_by_objective(Objective):
            candidates[f"{feat.key}_pred"] = np.nan
            candidates[f"{feat.key}_sd"] = np.nan
            candidates[f"{feat.key}_des"] = np.nan
        return candidates

    @property
    def input_preprocessing_specs(self):
        return {}

    def has_sufficient_experiments(
        self,
    ) -> bool:
        return len(self.experiments) >= 3


STRATEGY_MAP: Dict[Type[data_models.Strategy], Type[Strategy]] = {
    DummyStrategyDataModel: DummyStrategy,
    DummyPredictiveStrategyDataModel: DummyPredictiveStrategy,
}


def map(data_model: data_models.Strategy) -> Strategy:
    cls = STRATEGY_MAP[data_model.__class__]
    return cls.from_spec(data_model=data_model)



"""
Tanimoto Kernel. Operates on representations including bit vectors e.g. Morgan/ECFP6 fingerprints count vectors e.g.
RDKit fragment features.
"""
# This code was copied from GAUCHE: https://github.com/leojklarner/gauche/blob/main/gauche/kernels/fingerprint_kernels/tanimoto_kernel.py

import torch

from bofire.kernels.fingerprint_kernels.base_fingerprint_kernel import BitKernel


class TanimotoKernel(BitKernel):
    r"""
    Computes a covariance matrix based on the Tanimoto kernel between inputs `x1` and `x2`:

    Formula:
        .. math::

        \begin{equation*}
        k_{\text{Tanimoto}}(\mathbf{x}, \mathbf{x'}) = \frac{\langle\mathbf{x},
        \mathbf{x'}\rangle}{\left\lVert\mathbf{x}\right\rVert^2 + \left\lVert\mathbf{x'}\right\rVert^2 -
        \langle\mathbf{x}, \mathbf{x'}\rangle}
        \end{equation*}

    This kernel does not have an `outputscale` parameter. To add a scaling parameter,
    decorate this kernel with a `gpytorch.test_kernels.ScaleKernel`.

    Example:
        >>> x = torch.randint(0, 2, (10, 5))
        >>> # Non-batch: Simple option
        >>> covar_module = gpytorch.kernels.ScaleKernel(TanimotoKernel())
        >>> covar = covar_module(x)  # Output: LazyTensor of size (10 x 10)
        >>>
        >>> batch_x = torch.randint(0, 2, (2, 10, 5))
        >>> # Batch: Simple option
        >>> covar_module = gpytorch.kernels.ScaleKernel(TanimotoKernel())
        >>> covar = covar_module(batch_x)  # Output: LazyTensor of size (2 x 10 x 10)
    """
    is_stationary = False
    has_lengthscale = False

    def __init__(self, **kwargs):
        super(TanimotoKernel, self).__init__(**kwargs)
        self.metric = "tanimoto"

    def forward(self, x1, x2, diag=False, **params):
        if diag:
            assert x1.size() == x2.size() and torch.equal(x1, x2)
            return torch.ones(
                *x1.shape[:-2], x1.shape[-2], dtype=x1.dtype, device=x1.device
            )
        else:
            return self.covar_dist(x1, x2, **params)



from bofire.data_models.domain.api import Constraints, Domain, Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from tests.bofire.data_models.specs.features import specs as features
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])
specs.add_valid(
    Domain,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(key="i1"),
                features.valid(ContinuousInput).obj(key="i2"),
                features.valid(ContinuousInput).obj(key="i3"),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(key="o1"),
                features.valid(ContinuousOutput).obj(key="o2"),
            ]
        ),
        "constraints": Constraints(),
    },
)



import pandas as pd
from pydantic.error_wrappers import ValidationError
from pydantic.types import PositiveInt

import bofire.data_models.strategies.api as data_models
from bofire.strategies.samplers.polytope import PolytopeSampler
from bofire.strategies.samplers.rejection import RejectionSampler
from bofire.strategies.strategy import Strategy


class RandomStrategy(Strategy):
    """Strategy for randomly selecting new candidates.

    Provides a baseline strategy for benchmarks or for generating initial candidates.
    Uses PolytopeSampler or RejectionSampler, depending on the constraints.
    """

    def __init__(
        self,
        data_model: data_models.RandomStrategy,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self._init_sampler()

    def _init_sampler(self) -> None:
        errors = []
        for dm, functional in [
            (data_models.PolytopeSampler, PolytopeSampler),
            (data_models.RejectionSampler, RejectionSampler),
        ]:
            try:
                data_model = dm(domain=self.domain)
                self.sampler = functional(data_model=data_model)
                break
            except ValidationError as err:
                errors.append(err)
        if self.sampler is None:
            raise Exception(errors)

    def has_sufficient_experiments(self) -> bool:
        return True

    def _ask(self, candidate_count: PositiveInt) -> pd.DataFrame:
        return self.sampler.ask(candidate_count)  # type: ignore



from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from scipy.stats import chi2  # type: ignore

import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs


class OutlierDetection(ABC):
    @abstractmethod
    def detect(self, experiments: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    @abstractmethod
    def inputs(self) -> Inputs:
        pass

    @property
    @abstractmethod
    def outputs(self) -> Outputs:
        pass


class IterativeTrimming(OutlierDetection):
    def __init__(self, data_model, **kwargs):
        self.alpha1 = data_model.alpha1
        self.alpha2 = data_model.alpha2
        self.nsh = data_model.nsh
        self.ncc = data_model.ncc
        self.nrw = data_model.nrw
        self.base_gp = data_model.base_gp
        self.surrogate = surrogates.map(self.base_gp)
        super().__init__()

    @property
    def inputs(self) -> Inputs:
        return self.base_gp.inputs

    @property
    def outputs(self) -> Outputs:
        return self.base_gp.outputs

    def detect(self, experiments: pd.DataFrame) -> pd.DataFrame:
        n = len(experiments)
        indices = experiments.index.to_numpy()
        p = 1
        if n * self.alpha1 - 0.5 <= 2:
            raise ValueError("The dataset is unreasonably small!")
        d_sq = None
        ix_old = None
        niter = 0
        for i in range(1 + self.nsh + self.ncc):
            if i == 0:
                # starting with the full sample
                ix_sub = slice(None)
                consistency = 1.0
            else:
                # reducing alpha from 1 to alpha1 gradually
                if i <= self.nsh:
                    alpha = self.alpha1 + (1 - self.alpha1) * (1 - i / (self.nsh + 1))
                else:
                    alpha = self.alpha1
                chi_sq = chi2(p).ppf(alpha)
                h = int(min(np.ceil(n * alpha - 0.5), n - 1))  # alpha <= (h+0.5)/n

                # XXX: might be buggy when there are identical data points
                # better to use argpartition! but may break ix_sub == ix_old.
                ix_sub = d_sq <= np.partition(d_sq, h)[h]  # alpha-quantile
                consistency = alpha / chi2(p + 2).cdf(chi_sq)

            # check convergence
            if (i > self.nsh + 1) and (ix_sub == ix_old).all():  # type: ignore
                break  # converged
            ix_old = ix_sub

            self.surrogate.fit(experiments[experiments.index.isin(indices[ix_sub])])  # type: ignore

            # make prediction
            pred = self.surrogate.predict(experiments)
            d_sq = (
                (
                    experiments[self.base_gp.outputs.get_keys()[0]]
                    - pred[self.base_gp.outputs.get_keys()[0] + "_pred"]
                )
                ** 2
                / pred[self.base_gp.outputs.get_keys()[0] + "_sd"] ** 2
            ).ravel()

            niter += 1
        for _ in range(self.nrw):
            alpha = self.alpha2
            chi_sq = chi2(p).ppf(alpha)

            # XXX: might be buggy when there are identical data points
            ix_sub = d_sq <= chi_sq * consistency  # type: ignore
            consistency = alpha / chi2(p + 2).cdf(chi_sq)

            # check convergence
            if (ix_sub == ix_old).all():
                break  # converged
            ix_old = ix_sub

        filtered_experiments = experiments.copy()
        filtered_experiments.loc[
            ~ix_sub,  # type: ignore
            f"valid_{self.base_gp.outputs.get_keys()[0]}",  # type: ignore
        ] = 0
        return filtered_experiments  # type: ignore



# import importlib
import random
import uuid

import bofire.data_models.features.api as features
from bofire.data_models.objectives.api import MaximizeObjective
from tests.bofire.data_models.specs.objectives import specs as objectives
from tests.bofire.data_models.specs.specs import Specs

# RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None

specs = Specs([])

specs.add_valid(
    features.DiscreteInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "values": [random.random(), random.random() + 3],
        "unit": random.choice(["°C", "mg", "mmol/l", None]),
    },
)

specs.add_invalid(
    features.DiscreteInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "values": [1.0],
        "unit": random.choice(["°C", "mg", "mmol/l", None]),
    },
)

specs.add_valid(
    features.ContinuousInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "bounds": (3, 5.3),
        "unit": random.choice(["°C", "mg", "mmol/l", None]),
        "stepsize": None,
    },
)
specs.add_valid(
    features.ContinuousDescriptorInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "bounds": (3, 5.3),
        "descriptors": ["d1", "d2"],
        "values": [1.0, 2.0],
        "unit": random.choice(["°C", "mg", "mmol/l", None]),
        "stepsize": None,
    },
)
specs.add_valid(
    features.CategoricalInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "categories": ["c1", "c2", "c3"],
        "allowed": [True, True, False],
    },
)
specs.add_valid(
    features.CategoricalDescriptorInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "categories": ["c1", "c2", "c3"],
        "allowed": [True, True, False],
        "descriptors": ["d1", "d2"],
        "values": [
            [1.0, 2.0],
            [3.0, 7.0],
            [5.0, 1.0],
        ],
    },
)
specs.add_valid(
    features.ContinuousOutput,
    lambda: {
        "key": str(uuid.uuid4()),
        "objective": objectives.valid(MaximizeObjective).typed_spec(),
        "unit": random.choice(["%", "area %", None]),
    },
)

specs.add_valid(
    features.CategoricalOutput,
    lambda: {
        "key": str(uuid.uuid4()),
        "categories": ["a", "b", "c"],
        "objective": [0.0, 1.0, 0.0],
    },
)
specs.add_valid(
    features.MolecularInput,
    lambda: {
        "key": str(uuid.uuid4()),
    },
)

specs.add_valid(
    features.CategoricalMolecularInput,
    lambda: {
        "key": str(uuid.uuid4()),
        "categories": [
            "CC(=O)Oc1ccccc1C(=O)O",
            "c1ccccc1",
            "[CH3][CH2][OH]",
            "N[C@](C)(F)C(=O)O",
        ],
        "allowed": [True, True, True, True],
    },
)



from typing import Union

from bofire.data_models.strategies.doe import DoEStrategy
from bofire.data_models.strategies.factorial import FactorialStrategy
from bofire.data_models.strategies.predictives.botorch import BotorchStrategy
from bofire.data_models.strategies.predictives.mobo import MoboStrategy
from bofire.data_models.strategies.predictives.multiobjective import (
    MultiobjectiveStrategy,
)
from bofire.data_models.strategies.predictives.predictive import PredictiveStrategy
from bofire.data_models.strategies.predictives.qehvi import QehviStrategy
from bofire.data_models.strategies.predictives.qnehvi import QnehviStrategy
from bofire.data_models.strategies.predictives.qparego import QparegoStrategy
from bofire.data_models.strategies.predictives.sobo import (
    AdditiveSoboStrategy,
    CustomSoboStrategy,
    MultiplicativeSoboStrategy,
    SoboStrategy,
)
from bofire.data_models.strategies.random import RandomStrategy
from bofire.data_models.strategies.samplers.polytope import PolytopeSampler
from bofire.data_models.strategies.samplers.rejection import RejectionSampler
from bofire.data_models.strategies.samplers.sampler import SamplerStrategy
from bofire.data_models.strategies.stepwise.conditions import (  # noqa: F401
    AlwaysTrueCondition,
    CombiCondition,
    NumberOfExperimentsCondition,
)
from bofire.data_models.strategies.stepwise.stepwise import (  # noqa: F401
    Step,
    StepwiseStrategy,
)
from bofire.data_models.strategies.strategy import Strategy

AbstractStrategy = Union[
    Strategy,
    BotorchStrategy,
    SamplerStrategy,
    PredictiveStrategy,
    MultiobjectiveStrategy,
]

AnyStrategy = Union[
    SoboStrategy,
    AdditiveSoboStrategy,
    MultiplicativeSoboStrategy,
    CustomSoboStrategy,
    QehviStrategy,
    QnehviStrategy,
    QparegoStrategy,
    PolytopeSampler,
    RejectionSampler,
    RandomStrategy,
    DoEStrategy,
    StepwiseStrategy,
    FactorialStrategy,
    MoboStrategy,
]

AnyPredictive = Union[
    SoboStrategy,
    AdditiveSoboStrategy,
    MultiplicativeSoboStrategy,
    CustomSoboStrategy,
    QehviStrategy,
    QnehviStrategy,
    QparegoStrategy,
    MoboStrategy,
]

AnySampler = Union[PolytopeSampler, RejectionSampler]


AnyCondition = Union[NumberOfExperimentsCondition, CombiCondition, AlwaysTrueCondition]



import gpytorch.priors
import pytest

import bofire.priors.api as priors
from bofire.data_models.priors.api import GammaPrior, NormalPrior


@pytest.mark.parametrize(
    "prior, expected_prior",
    [
        (GammaPrior(concentration=2.0, rate=0.2), gpytorch.priors.GammaPrior),
        (NormalPrior(loc=0, scale=0.5), gpytorch.priors.NormalPrior),
    ],
)
def test_map(prior, expected_prior):
    gprior = priors.map(prior)
    assert isinstance(gprior, expected_prior)
    for key, value in prior.dict().items():
        if key == "type":
            continue
        assert value == getattr(gprior, key)



import json
import os
from copy import deepcopy
from typing import Callable, List, Optional, Protocol, Tuple, Union

import numpy as np
import pandas as pd
from multiprocess.pool import Pool
from tqdm import tqdm

from bofire.benchmarks.api import Benchmark
from bofire.data_models.domain.api import Domain
from bofire.strategies.api import Strategy


class StrategyFactory(Protocol):
    def __call__(self, domain: Domain) -> Strategy:
        ...


def _single_run(
    run_idx: int,
    benchmark: Benchmark,
    strategy_factory: StrategyFactory,
    n_iterations: int,
    metric: Callable[[Domain, pd.DataFrame], float],
    n_candidates_per_proposals: int,
    safe_intervall: int,
    initial_sampler: Optional[
        Union[Callable[[Domain], pd.DataFrame], pd.DataFrame]
    ] = None,
) -> Tuple[pd.DataFrame, pd.Series]:
    def autosafe_results(benchmark):
        """Safes results into a .json file to prevent data loss during time-expensive optimization runs.
        Autosave should operate every 10 iterations.

        Args:
            benchmark: Benchmark function that is suposed be evaluated.
        """

        benchmark_name = benchmark.__class__.__name__
        # Create a folder for autosaves, if not already exists.
        if not os.path.exists("bofire_autosaves/" + benchmark_name):
            os.makedirs("bofire_autosaves/" + benchmark_name)

        filename = (
            "bofire_autosaves/" + benchmark_name + "/run" + str(run_idx) + ".json"
        )
        parsed_domain = benchmark.domain.json()
        with open(filename, "w") as file:
            json.dump(parsed_domain, file)

    strategy = strategy_factory(domain=benchmark.domain)
    # sample initial values and tell
    if initial_sampler is not None:
        if isinstance(initial_sampler, Callable):
            X = initial_sampler(benchmark.domain)
            XY = benchmark.f(X, return_complete=True)
        else:
            XY = initial_sampler
        strategy.tell(XY)

    metric_values = np.zeros(n_iterations)
    pbar = tqdm(range(n_iterations), position=run_idx)
    for i in pbar:
        X = strategy.ask(candidate_count=n_candidates_per_proposals)
        X = X[benchmark.domain.inputs.get_keys()]
        Y = benchmark.f(X)
        XY = pd.concat([X, Y], axis=1)
        # pd.concat() changes datatype of str to np.int32 if column contains whole numbers.
        # colum needs to be converted back to str to be added to the benchmark domain.
        strategy.tell(XY)
        metric_values[i] = metric(strategy.domain, strategy.experiments)  # type: ignore
        pbar.set_description(
            f"run {run_idx:02d} with current best {metric_values[i]:0.3f}"
        )
        if (i + 1) % safe_intervall == 0:
            autosafe_results(benchmark=benchmark)
    return strategy.experiments, pd.Series(metric_values)


def run(
    benchmark: Benchmark,
    strategy_factory: StrategyFactory,
    n_iterations: int,
    metric: Callable[[Domain, pd.DataFrame], float],
    initial_sampler: Optional[Callable[[Domain], pd.DataFrame]] = None,
    n_candidates_per_proposal: int = 1,
    n_runs: int = 5,
    n_procs: int = 5,
    safe_intervall: int = 1000,
) -> List[Tuple[pd.DataFrame, pd.Series]]:
    """Run a benchmark problem several times in parallel

    Args:
        benchmark: problem to be benchmarked
        strategy_factory: creates the strategy to be benchmarked on the benchmark problem
        n_iterations: number of times the strategy is asked
        metric: measure of success, e.g, best value found so far for single objective or
                hypervolume for multi-objective
        initial_sampler: Creates initial data
        n_candidates: also known as batch size, number of proposals made at once by the strategy
        n_runs: number of runs
        n_procs: number of parallel processes to execute the runs

    Returns:
        per run, a tuple with the benchmark object containing the proposed data and metric values
    """

    def make_args(run_idx: int):
        return (
            run_idx,
            deepcopy(benchmark),
            strategy_factory,
            n_iterations,
            metric,
            n_candidates_per_proposal,
            safe_intervall,
            initial_sampler,
        )

    if n_procs == 1:
        results = [_single_run(*make_args(i)) for i in range(n_runs)]
    else:
        p = Pool(min(n_procs, n_runs))
        results = [p.apply_async(_single_run, make_args(i)) for i in range(n_runs)]
        results = [r.get() for r in results]
    return results



from typing import Literal, Type

from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
)
from bofire.data_models.strategies.strategy import Strategy


class FactorialStrategy(Strategy):
    type: Literal["FactorialStrategy"] = "FactorialStrategy"

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Feature]) -> bool:
        return False

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            CategoricalInput,
            DiscreteInput,
            CategoricalDescriptorInput,
            ContinuousOutput,
        ]



from typing import Union

from bofire.benchmarks.aspen_benchmark import Aspen_benchmark
from bofire.benchmarks.benchmark import Benchmark, GenericBenchmark
from bofire.benchmarks.hyperopt import Hyperopt
from bofire.benchmarks.multi import C2DTLZ2, DTLZ2, ZDT1, CrossCoupling, SnarBenchmark
from bofire.benchmarks.single import Ackley, Branin, Branin30, Hartmann, Himmelblau

AnyMultiBenchmark = Union[C2DTLZ2, DTLZ2, ZDT1, CrossCoupling, SnarBenchmark]
AnySingleBenchmark = Union[Ackley, Branin, Branin30, Hartmann, Himmelblau]



import pytest
import torch
import torch.nn as nn
from botorch.models.transforms.input import InputStandardize, Normalize
from botorch.models.transforms.outcome import Standardize
from pandas.testing import assert_frame_equal

import bofire.surrogates.api as surrogates
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.surrogates.api import MLPEnsemble, ScalerEnum
from bofire.surrogates.mlp import MLP, RegressionDataSet, _MLPEnsemble, fit_mlp
from bofire.utils.torch_tools import tkwargs


@pytest.mark.parametrize(
    "activation, expected",
    [
        ["relu", nn.modules.activation.ReLU],
        ["logistic", nn.modules.activation.Sigmoid],
        ["tanh", nn.modules.activation.Tanh],
    ],
)
def test_mlp_activation(activation, expected):
    mlp = MLP(input_size=2, output_size=1, activation=activation)
    assert isinstance(mlp.layers[1], expected)


def test_mlp_activation_invalid():
    with pytest.raises(ValueError):
        MLP(input_size=2, output_size=1, activation="mama")


@pytest.mark.parametrize("output_size", [1, 2])
def test_mlp_input_size(output_size):
    mlp = MLP(input_size=2, output_size=output_size)
    assert mlp.layers[-1].out_features == output_size


def test_mlp_hidden_layer_sizes():
    mlp = MLP(input_size=2, output_size=1, hidden_layer_sizes=(8, 4, 2))
    assert len(mlp.layers) == 7
    assert mlp.layers[0].in_features == 2
    assert mlp.layers[0].out_features == 8
    assert mlp.layers[2].in_features == 8
    assert mlp.layers[2].out_features == 4
    assert mlp.layers[4].in_features == 4
    assert mlp.layers[4].out_features == 2
    assert mlp.layers[6].in_features == 2
    assert mlp.layers[6].out_features == 1
    assert isinstance(mlp.layers[1], nn.modules.activation.ReLU)
    assert isinstance(mlp.layers[3], nn.modules.activation.ReLU)
    assert isinstance(mlp.layers[5], nn.modules.activation.ReLU)


def test_mlp_dropout():
    mlp = MLP(input_size=2, output_size=1, hidden_layer_sizes=(8, 4, 2), dropout=0.2)
    assert len(mlp.layers) == 10
    assert mlp.layers[0].in_features == 2
    assert mlp.layers[0].out_features == 8
    assert isinstance(mlp.layers[1], nn.modules.activation.ReLU)
    assert isinstance(mlp.layers[2], nn.modules.Dropout)
    assert mlp.layers[2].p == 0.2
    assert mlp.layers[3].in_features == 8
    assert isinstance(mlp.layers[4], nn.modules.activation.ReLU)
    assert isinstance(mlp.layers[5], nn.modules.Dropout)
    assert mlp.layers[3].out_features == 4
    assert mlp.layers[6].in_features == 4
    assert mlp.layers[6].out_features == 2
    assert isinstance(mlp.layers[7], nn.modules.activation.ReLU)
    assert isinstance(mlp.layers[8], nn.modules.Dropout)
    assert mlp.layers[9].in_features == 2
    assert mlp.layers[9].out_features == 1


def test_RegressionDataSet():
    X = torch.randn(10, 3)
    y = torch.randn(10, 1)
    dset = RegressionDataSet(X=X, y=y)
    assert len(dset) == 10
    xi, yi = dset[3]
    assert torch.allclose(xi, X[3].to(**tkwargs))
    assert torch.allclose(yi, y[3].to(**tkwargs))


@pytest.mark.parametrize(
    "mlp, weight_decay, n_epoches, lr, shuffle",
    [
        [
            MLP(input_size=2, output_size=1, hidden_layer_sizes=(8, 4, 2), dropout=0.2),
            0,
            10,
            1e-4,
            True,
        ],
        [
            MLP(input_size=2, output_size=1, hidden_layer_sizes=(8, 4, 2), dropout=0.2),
            0.1,
            5,
            1e-3,
            False,
        ],
    ],
)
def test_fit_mlp(mlp, weight_decay, n_epoches, lr, shuffle):
    benchmark = Himmelblau()
    samples = benchmark.domain.inputs.sample(10)
    experiments = benchmark.f(candidates=samples, return_complete=True)
    X, y = torch.from_numpy(experiments[["x_1", "x_2"]].values), torch.from_numpy(
        experiments[["y"]].values
    )
    dset = RegressionDataSet(X=X, y=y)
    fit_mlp(
        mlp=mlp,
        dataset=dset,
        weight_decay=weight_decay,
        n_epoches=n_epoches,
        lr=lr,
        shuffle=shuffle,
    )


def test_mlp_ensemble_no_mls():
    with pytest.raises(ValueError):
        _MLPEnsemble(mlps=[])


def test_mlp_ensemble_not_matching_models():
    mlp1 = MLP(input_size=2, output_size=1)
    mlp2 = MLP(input_size=3, output_size=1)
    with pytest.raises(AssertionError):
        _MLPEnsemble(mlps=[mlp1, mlp2])
    mlp1 = MLP(input_size=3, output_size=1)
    mlp2 = MLP(input_size=3, output_size=2)
    with pytest.raises(AssertionError):
        _MLPEnsemble(mlps=[mlp1, mlp2])


def test_mlp_ensemble_num_outputs():
    mlp1 = MLP(input_size=3, output_size=1)
    mlp2 = MLP(input_size=3, output_size=1)
    ens = _MLPEnsemble(mlps=[mlp1, mlp2])
    assert ens.num_outputs == 1


def test_mlp_ensemble_forward():
    bench = Himmelblau()
    samples = bench.domain.inputs.sample(10)
    experiments = bench.f(samples, return_complete=True)
    mlp1 = MLP(input_size=2, output_size=1)
    mlp2 = MLP(input_size=2, output_size=1)
    ens = _MLPEnsemble(mlps=[mlp1, mlp2])
    pred = ens.forward(torch.from_numpy(experiments[["x_1", "x_2"]].values))
    assert pred.shape == torch.Size((2, 10, 1))
    # test with batches
    batch = torch.from_numpy(experiments[["x_1", "x_2"]].values).unsqueeze(0)
    pred = ens.forward(batch)
    assert pred.shape == torch.Size((1, 2, 10, 1))


@pytest.mark.parametrize(
    "scaler, output_scaler",
    [
        [ScalerEnum.NORMALIZE, ScalerEnum.IDENTITY],
        [ScalerEnum.STANDARDIZE, ScalerEnum.STANDARDIZE],
        [ScalerEnum.IDENTITY, ScalerEnum.STANDARDIZE],
    ],
)
def test_mlp_ensemble_fit(scaler, output_scaler):
    bench = Himmelblau()
    samples = bench.domain.inputs.sample(10)
    experiments = bench.f(samples, return_complete=True)
    ens = MLPEnsemble(
        inputs=bench.domain.inputs,
        outputs=bench.domain.outputs,
        n_estimators=2,
        n_epochs=5,
        scaler=scaler,
        output_scaler=output_scaler,
    )
    surrogate = surrogates.map(ens)

    surrogate.fit(experiments=experiments)

    if scaler == ScalerEnum.NORMALIZE:
        assert isinstance(surrogate.model.input_transform, Normalize)
    elif scaler == ScalerEnum.STANDARDIZE:
        assert isinstance(surrogate.model.input_transform, InputStandardize)
    else:
        with pytest.raises(AttributeError):
            assert surrogate.model.input_transform is None

    if output_scaler == ScalerEnum.STANDARDIZE:
        assert isinstance(surrogate.model.outcome_transform, Standardize)
    elif output_scaler == ScalerEnum.IDENTITY:
        assert not hasattr(surrogate.model, "outcome_transform")

    preds = surrogate.predict(experiments)
    dump = surrogate.dumps()
    surrogate2 = surrogates.map(ens)
    surrogate2.loads(dump)
    preds2 = surrogate2.predict(experiments)
    assert_frame_equal(preds, preds2)


@pytest.mark.parametrize(
    "scaler", [ScalerEnum.NORMALIZE, ScalerEnum.STANDARDIZE, ScalerEnum.IDENTITY]
)
def test_mlp_ensemble_fit_categorical(scaler):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments.loc[experiments.x_cat == "mama", "y"] *= 5.0
    experiments.loc[experiments.x_cat == "papa", "y"] /= 2.0
    experiments["valid_y"] = 1

    ens = MLPEnsemble(
        inputs=inputs,
        outputs=outputs,
        n_estimators=2,
        n_epochs=5,
        scaler=scaler,
    )
    surrogate = surrogates.map(ens)
    surrogate.fit(experiments=experiments)
    if scaler == ScalerEnum.NORMALIZE:
        assert isinstance(surrogate.model.input_transform, Normalize)
        assert torch.eq(
            surrogate.model.input_transform.indices,
            torch.tensor([0, 1], dtype=torch.int64),
        ).all()
    elif scaler == ScalerEnum.STANDARDIZE:
        assert isinstance(surrogate.model.input_transform, InputStandardize)
        assert torch.eq(
            surrogate.model.input_transform.indices,
            torch.tensor([0, 1], dtype=torch.int64),
        ).all()
    else:
        with pytest.raises(AttributeError):
            assert surrogate.model.input_transform is None
    preds = surrogate.predict(experiments)
    dump = surrogate.dumps()
    surrogate2 = surrogates.map(ens)
    surrogate2.loads(dump)
    preds2 = surrogate2.predict(experiments)
    assert_frame_equal(preds, preds2)



from importlib.util import find_spec

import numpy as np
import pandas as pd
import pytest
import torch
from botorch.models import MixedSingleTaskGP, SingleTaskGP
from botorch.models.deterministic import DeterministicModel
from botorch.models.transforms.input import (
    ChainedInputTransform,
    FilterFeatures,
    InputStandardize,
    Normalize,
    OneHotToNumeric,
)
from pandas.testing import assert_frame_equal
from torch import Tensor

import bofire.data_models.surrogates.api as data_models
import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.surrogates.api import ScalerEnum
from bofire.surrogates.api import BotorchSurrogates
from bofire.surrogates.random_forest import _RandomForest
from bofire.utils.torch_tools import tkwargs

CLOUDPICKLE_NOT_AVAILABLE = find_spec("cloudpickle") is None


@pytest.mark.parametrize(
    "modelclass",
    [(data_models.SingleTaskGPSurrogate), (data_models.MixedSingleTaskGPSurrogate)],
)
def test_BotorchModel_validate_input_preprocessing_steps(modelclass):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(5)
        ]
        + [
            CategoricalInput(key="x_cat", categories=["mama", "papa"]),
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            ),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    data_model = modelclass(
        inputs=inputs,
        outputs=outputs,
    )
    surrogate = surrogates.map(data_model)
    assert surrogate.input_preprocessing_specs == {
        "x_cat": CategoricalEncodingEnum.ONE_HOT,
        "cat": CategoricalEncodingEnum.DESCRIPTOR,
    }
    # test that it can also handle incomplete specs
    data_model = modelclass(
        inputs=inputs,
        outputs=outputs,
        input_preprocessing_specs={"x_cat": CategoricalEncodingEnum.ONE_HOT},
    )
    surrogate = surrogates.map(data_model)
    assert surrogate.input_preprocessing_specs == {
        "x_cat": CategoricalEncodingEnum.ONE_HOT,
        "cat": CategoricalEncodingEnum.DESCRIPTOR,
    }


@pytest.mark.parametrize(
    "modelclass, input_preprocessing_specs",
    [
        (
            data_models.SingleTaskGPSurrogate,
            {
                "x_cat": CategoricalEncodingEnum.DUMMY,
                "cat": CategoricalEncodingEnum.ORDINAL,
            },
        ),
        (
            data_models.MixedSingleTaskGPSurrogate,
            {
                "x_cat": CategoricalEncodingEnum.ORDINAL,
                "cat": CategoricalEncodingEnum.ONE_HOT,
            },
        ),
    ],
)
def test_BotorchModel_validate_invalid_input_preprocessing_steps(
    modelclass, input_preprocessing_specs
):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(5)
        ]
        + [
            CategoricalInput(key="x_cat", categories=["mama", "papa"]),
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            ),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    with pytest.raises(ValueError):
        modelclass(
            inputs=inputs,
            outputs=outputs,
            input_preprocessing_specs=input_preprocessing_specs,
        )


def test_BotorchSurrogates_invalid_outputs():
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(3)]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
    )
    data_model2 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(2)]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
    )
    with pytest.raises(ValueError):
        data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])


def test_BotorchSurrogates_invalid_inputs():
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(3)]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
    )
    data_model2 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[ContinuousInput(key=f"x_{i+1}", bounds=(-4, 4)) for i in range(2)]
            + [CategoricalInput(key="x_3", categories=["apple", "banana"])]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
    )
    with pytest.raises(ValueError):
        data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])


def test_BotorchSurrogates_invalid_preprocessing():
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(3)
            ]
            + [
                CategoricalDescriptorInput(
                    key="cat",
                    categories=["apple", "banana"],
                    descriptors=["length", "width"],
                    values=[[1, 2], [3, 4]],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
    )
    data_model2 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(2)
            ]
            + [
                CategoricalDescriptorInput(
                    key="cat",
                    categories=["apple", "banana"],
                    descriptors=["length", "width"],
                    values=[[1, 2], [3, 4]],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y2")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat": CategoricalEncodingEnum.DESCRIPTOR},
    )
    with pytest.raises(ValueError):
        data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])


@pytest.mark.parametrize(
    "surrogate_list",
    [
        (
            [
                data_models.SingleTaskGPSurrogate(
                    inputs=Inputs(
                        features=[
                            ContinuousInput(
                                key=f"x_{i+1}",
                                bounds=(-4, 4),
                            )
                            for i in range(3)
                        ]
                        + [
                            CategoricalDescriptorInput(
                                key="cat",
                                categories=["apple", "banana"],
                                descriptors=["length", "width"],
                                values=[[1, 2], [3, 4]],
                            )
                        ]
                    ),
                    outputs=Outputs(features=[ContinuousOutput(key="y")]),
                    scaler=ScalerEnum.NORMALIZE,
                    input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
                ),
                data_models.SingleTaskGPSurrogate(
                    inputs=Inputs(
                        features=[
                            ContinuousInput(
                                key=f"x_{i+1}",
                                bounds=(-4, 4),
                            )
                            for i in range(2)
                        ]
                        + [
                            CategoricalDescriptorInput(
                                key="cat",
                                categories=["apple", "banana"],
                                descriptors=["length", "width"],
                                values=[[1, 2], [3, 4]],
                            )
                        ]
                    ),
                    outputs=Outputs(
                        features=[
                            ContinuousOutput(key="y2"),
                            ContinuousOutput(key="y3"),
                        ]
                    ),
                    scaler=ScalerEnum.NORMALIZE,
                    input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
                ),
            ]
        )
    ],
)
def test_botorch_models_invalid_number_of_outputs(surrogate_list):
    with pytest.raises(ValueError):
        data_models.BotorchSurrogates(surrogates=surrogate_list)


@pytest.mark.parametrize(
    "surrogate_list",
    [
        (
            [
                data_models.SingleTaskGPSurrogate(
                    inputs=Inputs(
                        features=[
                            ContinuousInput(
                                key=f"x_{i+1}",
                                bounds=(-4, 4),
                            )
                            for i in range(3)
                        ]
                        + [
                            CategoricalDescriptorInput(
                                key="cat",
                                categories=["apple", "banana"],
                                descriptors=["length", "width"],
                                values=[[1, 2], [3, 4]],
                            )
                        ]
                    ),
                    outputs=Outputs(features=[ContinuousOutput(key="y")]),
                    scaler=ScalerEnum.NORMALIZE,
                    input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
                ),
                data_models.SingleTaskGPSurrogate(
                    inputs=Inputs(
                        features=[
                            ContinuousInput(
                                key=f"x_{i+1}",
                                bounds=(-4, 4),
                            )
                            for i in range(2)
                        ]
                        + [
                            CategoricalDescriptorInput(
                                key="cat",
                                categories=["apple", "banana"],
                                descriptors=["length", "width"],
                                values=[[1, 2], [3, 4]],
                            )
                        ]
                    ),
                    outputs=Outputs(features=[ContinuousOutput(key="y2")]),
                    scaler=ScalerEnum.NORMALIZE,
                    input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
                ),
            ]
        )
    ],
)
def test_botorch_models_valid(surrogate_list):
    data_model = data_models.BotorchSurrogates(surrogates=surrogate_list)
    BotorchSurrogates(data_model=data_model)


def test_botorch_models_check_compatibility():
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(3)
            ]
            + [
                CategoricalDescriptorInput(
                    key="cat",
                    categories=["apple", "banana"],
                    descriptors=["length", "width"],
                    values=[[1, 2], [3, 4]],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
    )
    data_model2 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(2)
            ]
            + [
                CategoricalDescriptorInput(
                    key="cat",
                    categories=["apple", "banana"],
                    descriptors=["length", "width"],
                    values=[[1, 2], [3, 4]],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y2")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat": CategoricalEncodingEnum.ONE_HOT},
    )
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    models = BotorchSurrogates(data_model=data_model)
    # check too less input features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check unused input features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(4)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check wrong input feature
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(4)
        ]
        + [
            ContinuousInput(
                key="cat",
                bounds=(-4, 4),
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check too less output features
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y")])
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check too many output features
    out = Outputs(
        features=[
            ContinuousOutput(key="y"),
            ContinuousOutput(key="y2"),
            ContinuousOutput(key="y3"),
        ]
    )
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check wrong output features
    out = Outputs(
        features=[
            ContinuousOutput(key="y"),
            ContinuousOutput(key="y3"),
        ]
    )
    with pytest.raises(ValueError):
        models._check_compability(inp, out)
    # check valid
    inp = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
        + [
            CategoricalDescriptorInput(
                key="cat",
                categories=["apple", "banana"],
                descriptors=["length", "width"],
                values=[[1, 2], [3, 4]],
            )
        ]
    )
    out = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    models._check_compability(inp, out)


def test_botorch_models_input_preprocessing_specs():
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(3)
            ]
            + [
                CategoricalDescriptorInput(
                    key="cat",
                    categories=["apple", "banana"],
                    descriptors=["length", "width"],
                    values=[[1, 2], [3, 4]],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat": CategoricalEncodingEnum.DESCRIPTOR},
    )
    data_model2 = data_models.SingleTaskGPSurrogate(
        inputs=Inputs(
            features=[
                ContinuousInput(
                    key=f"x_{i+1}",
                    bounds=(-4, 4),
                )
                for i in range(2)
            ]
            + [
                CategoricalInput(
                    key="cat2",
                    categories=["lotta", "sarah"],
                )
            ]
        ),
        outputs=Outputs(features=[ContinuousOutput(key="y2")]),
        scaler=ScalerEnum.NORMALIZE,
        input_preprocessing_specs={"cat2": CategoricalEncodingEnum.ONE_HOT},
    )
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    surrogate = BotorchSurrogates(data_model=data_model)
    assert surrogate.input_preprocessing_specs == {
        "cat": CategoricalEncodingEnum.DESCRIPTOR,
        "cat2": CategoricalEncodingEnum.ONE_HOT,
    }


def test_botorch_models_invalid_compatibilize():
    # model 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments1 = inputs.sample(n=10)
    experiments1.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments1["valid_y"] = 1
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        scaler=ScalerEnum.NORMALIZE,
    )
    data_model2 = data_models.EmpiricalSurrogate(
        inputs=inputs,
        outputs=Outputs(features=[ContinuousOutput(key="y2")]),
    )
    # create models
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    botorch_surrogates = BotorchSurrogates(data_model=data_model)
    botorch_surrogates.fit(experiments=experiments1)
    # try to make compatible
    with pytest.raises(ValueError):
        botorch_surrogates.compatibilize(
            inputs=inputs,
            outputs=Outputs(
                features=[ContinuousOutput(key="y2"), ContinuousOutput(key="y")]
            ),
        )


def test_botorch_models_fit_and_compatibilize():
    # model 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments1 = inputs.sample(n=10)
    experiments1.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments1["valid_y"] = 1
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        scaler=ScalerEnum.NORMALIZE,
    )
    # model 2
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y2")])
    experiments2 = pd.concat(
        [experiments1, inputs.get_by_key("x_cat").sample(10)], axis=1
    )
    experiments2.eval("y2=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments2.loc[experiments2.x_cat == "mama", "y2"] *= 5.0
    experiments2.loc[experiments2.x_cat == "papa", "y2"] /= 2.0
    experiments2["valid_y2"] = 1
    data_model2 = data_models.MixedSingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        input_preprocessing_specs={"x_cat": CategoricalEncodingEnum.ONE_HOT},
        scaler=ScalerEnum.STANDARDIZE,
    )
    # create models
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    botorch_surrogates = BotorchSurrogates(data_model=data_model)
    # unite experiments
    experiments = pd.concat(
        [experiments1, experiments2[["x_cat", "y2", "valid_y2"]]],
        axis=1,
        ignore_index=False,
    )
    # fit the models
    botorch_surrogates.fit(experiments=experiments)
    assert botorch_surrogates.surrogates[0].is_compatibilized is False
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # make and store predictions for later comparison
    preds1 = botorch_surrogates.surrogates[0].predict(experiments1)
    preds2 = botorch_surrogates.surrogates[1].predict(experiments2)
    # make compatible
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    combined = botorch_surrogates.compatibilize(inputs=inputs, outputs=outputs)
    assert botorch_surrogates.surrogates[0].is_compatibilized is True
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # check combined
    assert isinstance(combined.models[0], SingleTaskGP)
    assert isinstance(combined.models[1], MixedSingleTaskGP)
    assert isinstance(combined.models[0].input_transform, ChainedInputTransform)
    assert isinstance(combined.models[0].input_transform.tcompatibilize, FilterFeatures)
    assert torch.eq(
        combined.models[0].input_transform.tcompatibilize.feature_indices,
        torch.tensor([0, 1], dtype=torch.int64),
    ).all()
    assert isinstance(combined.models[0].input_transform.tf2, Normalize)
    assert isinstance(combined.models[1].input_transform, ChainedInputTransform)
    assert isinstance(combined.models[1].input_transform.tf1, InputStandardize)
    assert isinstance(combined.models[1].input_transform.tf2, OneHotToNumeric)
    # check predictions
    # transform experiments to torch
    trX = inputs.transform(
        experiments=experiments, specs={"x_cat": CategoricalEncodingEnum.ONE_HOT}
    )
    X = torch.from_numpy(trX.values).to(**tkwargs)
    with torch.no_grad():
        preds = combined.posterior(X).mean.detach().numpy()

    assert np.allclose(preds1.y_pred.values, preds[:, 0])
    assert np.allclose(preds2.y2_pred.values, preds[:, 1])
    ## now decompatibilize the models again
    botorch_surrogates.surrogates[0].decompatibilize()
    botorch_surrogates.surrogates[1].decompatibilize()
    assert botorch_surrogates.surrogates[0].is_compatibilized is False
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # check again the predictions
    preds11 = botorch_surrogates.surrogates[0].predict(experiments1)
    preds22 = botorch_surrogates.surrogates[1].predict(experiments2)
    assert_frame_equal(preds1, preds11)
    assert_frame_equal(preds2, preds22)
    assert isinstance(botorch_surrogates.surrogates[0].model.input_transform, Normalize)


def test_botorch_models_rf_fit_and_compatibilize():
    # model 1
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments1 = inputs.sample(n=10)
    experiments1.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments1["valid_y"] = 1
    data_model1 = data_models.SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        scaler=ScalerEnum.NORMALIZE,
    )
    # model 2
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y2")])
    experiments2 = pd.concat(
        [experiments1, inputs.get_by_key("x_cat").sample(10)], axis=1
    )
    experiments2.eval("y2=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments2.loc[experiments2.x_cat == "mama", "y2"] *= 5.0
    experiments2.loc[experiments2.x_cat == "papa", "y2"] /= 2.0
    experiments2["valid_y2"] = 1
    data_model2 = data_models.RandomForestSurrogate(
        inputs=inputs,
        outputs=outputs,
        input_preprocessing_specs={"x_cat": CategoricalEncodingEnum.ONE_HOT},
    )
    # create models
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    botorch_surrogates = BotorchSurrogates(data_model=data_model)
    # unite experiments
    experiments = pd.concat(
        [experiments1, experiments2[["x_cat", "y2", "valid_y2"]]],
        axis=1,
        ignore_index=False,
    )
    # fit the models
    botorch_surrogates.fit(experiments=experiments)
    assert botorch_surrogates.surrogates[0].is_compatibilized is False
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # make and store predictions for later comparison
    preds1 = botorch_surrogates.surrogates[0].predict(experiments1)
    preds2 = botorch_surrogates.surrogates[1].predict(experiments2)
    # make compatible
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    combined = botorch_surrogates.compatibilize(inputs=inputs, outputs=outputs)
    assert botorch_surrogates.surrogates[0].is_compatibilized is True
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # check combined
    assert isinstance(combined.models[0], SingleTaskGP)
    assert isinstance(combined.models[1], _RandomForest)
    assert isinstance(combined.models[0].input_transform, ChainedInputTransform)
    assert isinstance(combined.models[0].input_transform.tcompatibilize, FilterFeatures)
    assert torch.eq(
        combined.models[0].input_transform.tcompatibilize.feature_indices,
        torch.tensor([0, 1], dtype=torch.int64),
    ).all()
    assert isinstance(combined.models[0].input_transform.tf2, Normalize)
    # check predictions
    # transform experiments to torch
    trX = inputs.transform(
        experiments=experiments, specs={"x_cat": CategoricalEncodingEnum.ONE_HOT}
    )
    X = torch.from_numpy(trX.values).to(**tkwargs)
    with torch.no_grad():
        preds = combined.posterior(X).mean.detach().numpy()

    assert np.allclose(preds1.y_pred.values, preds[:, 0])
    assert np.allclose(preds2.y2_pred.values, preds[:, 1])
    ## now decompatibilize the models again
    botorch_surrogates.surrogates[0].decompatibilize()
    botorch_surrogates.surrogates[1].decompatibilize()
    assert botorch_surrogates.surrogates[0].is_compatibilized is False
    assert botorch_surrogates.surrogates[1].is_compatibilized is False
    # check again the predictions
    preds11 = botorch_surrogates.surrogates[0].predict(experiments1)
    preds22 = botorch_surrogates.surrogates[1].predict(experiments2)
    assert_frame_equal(preds1, preds11)
    assert_frame_equal(preds2, preds22)
    assert isinstance(botorch_surrogates.surrogates[0].model.input_transform, Normalize)


class HimmelblauModel(DeterministicModel):
    def __init__(self):
        super().__init__()
        self._num_outputs = 1

    def forward(self, X: Tensor) -> Tensor:
        return (
            (X[..., 0] ** 2 + X[..., 1] - 11.0) ** 2
            + (X[..., 0] + X[..., 1] ** 2 - 7.0) ** 2
        ).unsqueeze(-1)


def test_empirical_model():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments1 = inputs.sample(n=10)
    experiments1.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments1["valid_y"] = 1
    data_model1 = data_models.EmpiricalSurrogate(inputs=inputs, outputs=outputs)
    surrogate1 = surrogates.map(data_model1)
    surrogate1.model = HimmelblauModel()
    # test prediction
    preds1 = surrogate1.predict(experiments1)
    assert np.allclose(experiments1.y.values, preds1.y_pred.values)
    # test usage of empirical model in a modellist
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y2")])
    experiments2 = pd.concat(
        [experiments1, inputs.get_by_key("x_cat").sample(10)], axis=1
    )
    experiments2.eval("y2=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments2.loc[experiments2.x_cat == "mama", "y2"] *= 5.0
    experiments2.loc[experiments2.x_cat == "papa", "y2"] /= 2.0
    experiments2["valid_y2"] = 1
    data_model2 = data_models.MixedSingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        input_preprocessing_specs={"x_cat": CategoricalEncodingEnum.ONE_HOT},
        scaler=ScalerEnum.STANDARDIZE,
    )
    # create models
    data_model = data_models.BotorchSurrogates(surrogates=[data_model1, data_model2])
    botorch_surrogates = BotorchSurrogates(data_model=data_model)
    # unite experiments
    experiments = pd.concat(
        [experiments1, experiments2[["x_cat", "y2", "valid_y2"]]],
        axis=1,
        ignore_index=False,
    )
    # fit the models
    botorch_surrogates.fit(experiments=experiments)
    botorch_surrogates.surrogates[0].model = HimmelblauModel()
    # make and store predictions for later comparison
    preds1 = botorch_surrogates.surrogates[0].predict(experiments1)
    preds2 = botorch_surrogates.surrogates[1].predict(experiments2)
    # make compatible
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y"), ContinuousOutput(key="y2")])
    combined = botorch_surrogates.compatibilize(inputs=inputs, outputs=outputs)
    # check combined
    assert isinstance(combined.models[0], DeterministicModel)
    assert isinstance(combined.models[1], MixedSingleTaskGP)
    assert isinstance(combined.models[0].input_transform, FilterFeatures)
    assert torch.eq(
        combined.models[0].input_transform.feature_indices,
        torch.tensor([0, 1], dtype=torch.int64),
    ).all()
    assert isinstance(combined.models[1].input_transform, ChainedInputTransform)
    assert isinstance(combined.models[1].input_transform.tf1, InputStandardize)
    assert isinstance(combined.models[1].input_transform.tf2, OneHotToNumeric)
    # check predictions
    # transform experiments to torch
    trX = inputs.transform(
        experiments=experiments, specs={"x_cat": CategoricalEncodingEnum.ONE_HOT}
    )
    X = torch.from_numpy(trX.values).to(**tkwargs)
    with torch.no_grad():
        preds = combined.posterior(X).mean.detach().numpy()

    assert np.allclose(preds1.y_pred.values, preds[:, 0])
    assert np.allclose(preds2.y2_pred.values, preds[:, 1])


@pytest.mark.skipif(CLOUDPICKLE_NOT_AVAILABLE, reason="requires cloudpickle")
def test_empirical_model_io():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    data_model = data_models.EmpiricalSurrogate(inputs=inputs, outputs=outputs)
    surrogate = surrogates.map(data_model)
    with pytest.raises(ValueError):
        surrogate.dumps()
    surrogate.model = HimmelblauModel()
    samples = inputs.sample(5)
    preds = surrogate.predict(samples)
    dump = surrogate.dumps()
    #
    data_model2 = data_models.EmpiricalSurrogate(inputs=inputs, outputs=outputs)
    surrogate2 = surrogates.map(data_model2)
    surrogate2.loads(dump)
    preds2 = surrogate2.predict(samples)
    assert_frame_equal(preds, preds2)
    # test with dump in datamodel
    data_model.dump = dump
    surrogate3 = surrogates.map(data_model)
    preds3 = surrogate3.predict(samples)
    assert_frame_equal(preds, preds3)



import math
from typing import Optional

import numpy as np
import pandas as pd
import torch
from botorch.test_functions import Hartmann as botorch_hartmann
from botorch.test_functions.synthetic import Branin as torchBranin
from pydantic.types import PositiveInt

from bofire.benchmarks.benchmark import Benchmark
from bofire.data_models.constraints.api import NChooseKConstraint
from bofire.data_models.domain.api import Constraints, Domain, Inputs, Outputs
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.data_models.objectives.api import MaximizeObjective, MinimizeObjective
from bofire.utils.torch_tools import tkwargs


class Ackley(Benchmark):
    """Ackley function for testing optimization algorithms
    Virtual experiment corresponds to a function evaluation.
    Examples
    --------
    >>> b = Ackley()
    >>> columns = [v.name for v in b.domain.variables]
    >>> values = [v.bounds[0]+0.1*(v.bounds[1]-v.bounds[0]) for v in b.domain.variables]
    >>> values = np.array(values)
    >>> conditions = DataSet(values, columns=columns)
    >>> results = b.run_experiments(conditions)
    Notes
    -----
    This function is the negated version of https://en.wikipedia.org/wiki/Ackley_function.
    """

    # @validator("validate_categoricals")
    # def validate_categoricals(cls, v, num_categoricals):
    #     if v and num_categoricals ==1:
    #         raise ValueError("num_categories  must be specified if categorical=True")
    #     return v

    def __init__(
        self,
        num_categories: PositiveInt = 3,
        categorical: bool = False,
        descriptor: bool = False,
        dim: PositiveInt = 2,
        lower: float = -32.768,
        upper: float = 32.768,
        best_possible_f: float = 0.0,
        evaluated_points: Optional[list] = None,
        **kwargs,
    ):
        """Initializes benchmark function of type Ackley.

        Args:
            num_categories (PositiveInt, optional): Number of categories. Defaults to 3.
            categorical (bool, optional): Use categorical inputs. Defaults to False.
            descriptor (bool, optional): Use descriptive inputs. Defaults to False.
            dim (PositiveInt, optional): Dimension. Defaults to 2.
            lower (float, optional): Upper boundary. Defaults to -32.768.
            upper (float, optional): Lower boundary. Defaults to 32.768.
            best_possible_f (float, optional): Best possible function value. Defaults to 0.0.
            evaluated_points (list, optional): Evaluated points. Defaults to [].
        """
        super().__init__(**kwargs)
        self.num_categories = num_categories
        self.categorical = categorical
        self.descriptor = descriptor
        self.dim = dim
        self.lower = lower
        self.upper = upper
        self.best_possible_f = best_possible_f
        if evaluated_points is None:
            evaluated_points = []
        self.evaluated_points = evaluated_points

        input_feature_list = []
        # Decision variables
        if self.categorical:
            input_feature_list.append(
                CategoricalInput(
                    key="category",
                    categories=[str(x) for x in range(self.num_categories)],
                )
            )

        if self.descriptor:
            input_feature_list.append(
                CategoricalDescriptorInput(
                    key="descriptor",
                    categories=[str(x) for x in range(self.num_categories)],
                    descriptors=["d1"],
                    values=[[x * 2] for x in range(self.num_categories)],
                )
            )

        # continuous input features
        for d in range(self.dim):
            input_feature_list.append(
                ContinuousInput(key=f"x_{d+1}", bounds=(self.lower, self.upper))
            )

        # Objective
        output_feature = ContinuousOutput(key="y", objective=MaximizeObjective(w=1))

        self._domain = Domain(
            inputs=Inputs(features=input_feature_list),
            outputs=Outputs(features=[output_feature]),
        )

    def _f(self, X: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """Evaluates benchmark function.

        Args:
            X (pd.DataFrame): Input values. Columns are x_1 and x_2

        Returns:
            pd.DataFrame: y values of the function. Columns are y and valid_y.
        """
        a = 20
        b = 0.2
        c = np.pi * 2
        x = np.array([X[f"x_{d+1}"] for d in range(self.dim)])

        c = np.zeros(len(X))
        d = np.zeros(len(X))
        n = self.dim

        if self.categorical:
            # c = pd.to_numeric(X["category"], downcast="float")
            c = X.loc[:, "category"].values.astype(np.float64)
        if self.descriptor:
            d = X.loc[:, "descriptor"].values.astype(np.float64)

        z = x + c + d

        term1 = -a * np.exp(-b * ((1 / n) * np.sum(z**2, axis=0)) ** 0.5)
        term2 = -np.exp((1 / n) * np.sum(np.cos(c * z), axis=0))
        term3 = a + np.exp(1)
        y = term1 + term2 + term3
        Y = pd.DataFrame({"y": y, "valid_y": 1})
        # save evaluated points for plotting
        self.evaluated_points.append(x.tolist())
        return Y

    def get_optima(self) -> pd.DataFrame:
        """Returns positions of optima of the benchmark function.

        Returns:
            pd.DataFrame: x values of optima. Colums are x_1, x_2, y and valid_y
        """
        x = np.zeros((1, self.dim))
        y = 0
        return pd.DataFrame(
            np.c_[x, y],
            columns=self.domain.inputs.get_keys() + self.domain.outputs.get_keys(),
        )


class Hartmann(Benchmark):
    def __init__(self, dim: int = 6, allowed_k: Optional[int] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self._domain = Domain(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i}", bounds=(0, 1)) for i in range(dim)
                ]
            ),
            outputs=Outputs(
                features=[ContinuousOutput(key="y", objective=MinimizeObjective())]
            ),
            constraints=Constraints(
                constraints=[
                    NChooseKConstraint(
                        features=[f"x_{i}" for i in range(dim)],
                        min_count=0,
                        max_count=allowed_k,
                        none_also_valid=True,
                    )
                ]
            )
            if allowed_k
            else Constraints(),
        )
        self._hartmann = botorch_hartmann(dim=dim)

    def get_optima(self) -> pd.DataFrame:
        if self.dim != 6:
            raise ValueError("Only available for dim==6.")
        if len(self.domain.constraints) > 0:
            raise ValueError("Not defined for NChooseK use case.")
        return pd.DataFrame(
            columns=[f"x_{i}" for i in range(self.dim)] + ["y"],
            data=[[0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573, -3.32237]],
        )

    @property
    def dim(self) -> int:
        return len(self.domain.inputs)

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "y": self._hartmann(
                    torch.from_numpy(
                        candidates[[f"x_{i}" for i in range(self.dim)]].values
                    )
                ),
                "valid_y": [1 for _ in range(len(candidates))],
            }
        )


class Branin(Benchmark):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._domain = Domain(
            inputs=Inputs(
                features=[
                    ContinuousInput(key="x_1", bounds=(-5.0, 10)),
                    ContinuousInput(key="x_2", bounds=(0.0, 15.0)),
                ]
            ),
            outputs=Outputs(
                features=[ContinuousOutput(key="y", objective=MinimizeObjective())]
            ),
        )
        self.branin = torchBranin().to(**tkwargs)

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        c = torch.from_numpy(candidates[self.domain.inputs.get_keys()].values).to(
            **tkwargs
        )
        return pd.DataFrame(
            {
                "y": self.branin(c).detach().numpy(),
                "valid_y": np.ones(len(candidates)),
            }
        )

    def get_optima(self) -> pd.DataFrame:
        return pd.DataFrame(
            data=np.array(
                [
                    [-math.pi, 12.275, 0.397887],
                    [math.pi, 2.275, 0.397887],
                    [9.42478, 2.475, 0.397887],
                ]
            ),
            columns=self.domain.inputs.get_keys() + self.domain.outputs.get_keys(),
        )


class Branin30(Benchmark):
    """Thirty dimensional Branin function in which only the first two dimensions are used to
    evaluate the actual Branin. Source: https://github.com/pytorch/botorch/blob/main/tutorials/saasbo.ipynb.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._domain = Domain(
            inputs=Inputs(
                features=[
                    ContinuousInput(key=f"x_{i+1:02d}", bounds=(0, 1))
                    for i in range(30)
                ]
            ),
            outputs=Outputs(
                features=[ContinuousOutput(key="y", objective=MinimizeObjective())]
            ),
        )
        self.branin = torchBranin().to(**tkwargs)

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        lb, ub = self.branin.bounds  # type: ignore
        c = torch.from_numpy(candidates[self.domain.inputs.get_keys()].values).to(
            **tkwargs
        )
        return pd.DataFrame(
            {
                "y": self.branin(lb + (ub - lb) * c[..., :2]).detach().numpy(),
                "valid_y": np.ones(len(candidates)),
            }
        )


class Himmelblau(Benchmark):
    """Himmelblau function for testing optimization algorithms
    Link to the definition: https://en.wikipedia.org/wiki/Himmelblau%27s_function
    """

    def __init__(self, use_constraints: bool = False, **kwargs):
        """Initialiszes class of type Himmelblau.

        Args:
            best_possible_f (float, optional): Not implemented yet. Defaults to 0.0.
            use_constraints (bool, optional): Whether constraints should be used or not (Not implemented yet.). Defaults to False.

        Raises:
            ValueError: As constraints are not implemeted yet, a True value for use_constraints yields a ValueError.
        """
        super().__init__(**kwargs)
        self.use_constraints = use_constraints
        inputs = []

        inputs.append(ContinuousInput(key="x_1", bounds=(-6, 6)))
        inputs.append(ContinuousInput(key="x_2", bounds=(-6, 6)))

        objective = MinimizeObjective(w=1.0)
        output_feature = ContinuousOutput(key="y", objective=objective)
        if self.use_constraints:
            raise ValueError("Not implemented yet!")
        self._domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=[output_feature]),
        )

    def _f(self, X: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """Evaluates benchmark function.

        Args:
            X (pd.DataFrame): Input values. Columns are x_1 and x_2

        Returns:
            pd.DataFrame: y values of the function. Columns are y and valid_y.
        """
        X_temp = X.eval(
            "y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=False
        )
        Y = pd.DataFrame({"y": X_temp["y"], "valid_y": 1})
        return Y

    def get_optima(self) -> pd.DataFrame:
        """Returns positions of optima of the benchmark function.

        Returns:
            pd.DataFrame: x values of optima. Colums are x_1 and x_2
        """
        x = np.array(
            [
                [3.0, 2.0],
                [-2.805118, 3.131312],
                [-3.779310, -3.283186],
                [3.584428, -1.848126],
            ]
        )
        y = np.zeros(4)
        return pd.DataFrame(
            np.c_[x, y],
            columns=self.domain.inputs.get_keys() + self.domain.outputs.get_keys(),
        )


class DiscreteHimmelblau(Himmelblau):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        inputs = []

        inputs.append(DiscreteInput(key="x_1", values=np.linspace(-6, 6, 20).tolist()))
        inputs.append(DiscreteInput(key="x_2", values=np.linspace(-6, 6, 20).tolist()))

        objective = MinimizeObjective(w=1.0)
        output_feature = ContinuousOutput(key="y", objective=objective)

        self._domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=[output_feature]),
        )


class _CategoricalDiscreteHimmelblau(Himmelblau):  # only used for testing
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        inputs = []

        inputs.append(DiscreteInput(key="x_1", values=np.linspace(-6, 6, 20).tolist()))
        inputs.append(DiscreteInput(key="x_2", values=np.linspace(-6, 6, 20).tolist()))
        inputs.append(CategoricalInput(key="x_3", categories=["a", "b", "c"]))

        objective = MinimizeObjective(w=1.0)
        output_feature = ContinuousOutput(key="y", objective=objective)

        self._domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=[output_feature]),
        )

    def get_optima(self):
        raise NotImplementedError()



from typing import Literal, Type, Union

from bofire.data_models.constraints.api import Constraint
from bofire.data_models.features.api import (
    Feature,
    MolecularInput,
)
from bofire.data_models.objectives.api import Objective
from bofire.data_models.strategies.strategy import Strategy


class DoEStrategy(Strategy):
    type: Literal["DoEStrategy"] = "DoEStrategy"
    formula: Union[
        Literal[
            "linear",
            "linear-and-quadratic",
            "linear-and-interactions",
            "fully-quadratic",
        ],
        str,
    ]
    optimization_strategy: Literal[
        "default",
        "exhaustive",
        "branch-and-bound",
        "partially-random",
        "relaxed",
        "iterative",
    ] = "default"

    verbose: bool = False

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        return True

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        if my_type in [MolecularInput]:
            return False
        return True

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        return True



import pytest

from bofire.data_models.domain.api import Inputs
from bofire.data_models.domain.features import ContinuousInput
from bofire.utils.doe import get_confounding_matrix

inputs = Inputs(
    features=[ContinuousInput(key=i, bounds=(0, 10)) for i in ["a", "b", "c"]]
)


design = inputs.sample(20)


@pytest.mark.parametrize(
    "powers, interactions",
    [([-1], [2]), ([2], [1]), ([2], [4])],
)
def test_get_confounding_matrix_invalid(powers, interactions):
    with pytest.raises(AssertionError):
        get_confounding_matrix(
            inputs=inputs,
            design=design,
            powers=powers,
            interactions=interactions,
        )


@pytest.mark.parametrize(
    "powers, interactions",
    [([2], [2]), ([3], [2]), ([3], [2, 3])],
)
def test_get_confounding_matrix_valid(powers, interactions):
    get_confounding_matrix(
        inputs=inputs,
        design=design,
        powers=powers,
        interactions=interactions,
    )



import pytest

from bofire.benchmarks.api import Himmelblau
from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.kernels.api import MaternKernel, RBFKernel
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate
from bofire.runners.api import hyperoptimize

benchmark = Himmelblau()
experiments = benchmark.f(benchmark.domain.inputs.sample(9), return_complete=True)


def test_hyperoptimize_warning():
    surrogate_data = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs,
        outputs=benchmark.domain.outputs,
        hyperconfig=None,
    )
    with pytest.warns(
        match="No hyperopt is possible as no hyperopt config is available. Returning initial config."
    ):
        opt_data, metrics = hyperoptimize(
            surrogate_data=surrogate_data, training_data=experiments, folds=3
        )
    assert opt_data == surrogate_data
    assert len(metrics) == 0
    assert set(metrics.columns) == {e.name for e in RegressionMetricsEnum}


@pytest.mark.parametrize("strategy", ["FactorialStrategy", "RandomStrategy"])
def test_hyperoptimize(strategy: str):
    surrogate_data = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs,
        outputs=benchmark.domain.outputs,
    )
    if strategy == "RandomStrategy":
        surrogate_data.hyperconfig.hyperstrategy = strategy
        surrogate_data.hyperconfig.n_iterations = 5

    opt_data, metrics = hyperoptimize(
        surrogate_data=surrogate_data, training_data=experiments, folds=3
    )
    if strategy == "RandomStrategy":
        assert len(metrics) == 5
    else:
        assert len(metrics) == 12

    assert set(metrics.columns) == set(
        [e.name for e in RegressionMetricsEnum]
        + surrogate_data.hyperconfig.domain.inputs.get_keys()
    )
    assert opt_data.kernel.base_kernel.ard == (metrics.iloc[0]["ard"] == "True")
    if metrics.iloc[0].kernel == "matern_1.5":
        assert isinstance(opt_data.kernel.base_kernel, MaternKernel)
        assert opt_data.kernel.base_kernel.nu == 1.5
    elif metrics.iloc[0].kernel == "matern_2.5":
        assert isinstance(opt_data.kernel.base_kernel, MaternKernel)
        assert opt_data.kernel.base_kernel.nu == 2.5
    else:
        assert isinstance(opt_data.kernel.base_kernel, RBFKernel)



from typing import Union

from bofire.data_models.kernels.aggregation import (
    AdditiveKernel,
    MultiplicativeKernel,
    ScaleKernel,
)
from bofire.data_models.kernels.categorical import (
    CategoricalKernel,
    HammondDistanceKernel,
)
from bofire.data_models.kernels.continuous import (
    ContinuousKernel,
    LinearKernel,
    MaternKernel,
    PolynomialKernel,
    RBFKernel,
)
from bofire.data_models.kernels.kernel import Kernel
from bofire.data_models.kernels.molecular import MolecularKernel, TanimotoKernel

AbstractKernel = Union[Kernel, CategoricalKernel, ContinuousKernel, MolecularKernel]

AnyContinuousKernel = Union[
    MaternKernel,
    LinearKernel,
    PolynomialKernel,
    RBFKernel,
]

AnyCategoricalKernal = HammondDistanceKernel

AnyMolecularKernel = TanimotoKernel

AnyKernel = Union[
    AdditiveKernel,
    MultiplicativeKernel,
    ScaleKernel,
    HammondDistanceKernel,
    LinearKernel,
    PolynomialKernel,
    MaternKernel,
    RBFKernel,
    TanimotoKernel,
]



from __future__ import annotations

import argparse
import os
import subprocess
import time
from pathlib import Path
from typing import Dict, Optional

import pandas as pd

# This script is based on the corresponding one from botorch: https://github.com/pytorch/botorch/blob/main/scripts/run_tutorials.py


def run_script(
    tutorial: Path, timeout_minutes: int = 20, env: Optional[Dict[str, str]] = None
):
    utils_path = {"PYTHONPATH": str(tutorial.parent)}
    if env is not None:
        env = {**os.environ, **env, **utils_path}
    else:
        env = {**os.environ, **utils_path}
    try:
        run_out = subprocess.run(
            ["papermill", str(tutorial.absolute()), "temp.ipynb"],  # , "|"
            capture_output=True,
            text=True,
            env=env,
            encoding="utf-8",
            timeout=timeout_minutes * 60,
        )
    except subprocess.TimeoutExpired:
        print(f"{tutorial} exceeded max. runtime ({timeout_minutes*60} s)... ")
        return None
    return run_out


def run_tutorials(
    name: Optional[str] = None,
    smoke_test=False,
) -> None:
    """
    Run each tutorial, print statements on how it ran, and write a data set as a csv
    to a directory.
    """

    timeout_minutes = 30 if smoke_test is False else 2

    print(f"Running Tutorials, smoke_test_flag = {smoke_test}")

    tutorial_dir = Path(os.getcwd()).joinpath("tutorials")
    num_runs = 0
    num_errors = 0

    tutorials = sorted(t for t in tutorial_dir.rglob("*.ipynb") if t.is_file)
    env = {"SMOKE_TEST": "True"} if smoke_test else None
    if name is not None:
        tutorials = [t for t in tutorials if t.name == name]
        if len(tutorials) == 0:
            raise RuntimeError(f"Specified tutorial {name} not found in directory.")

    df = pd.DataFrame(
        {
            "name": [t.name for t in tutorials],
            "ran_successfully": False,
            "message": "",
            "runtime": float("nan"),
        }
    ).set_index("name")

    # ToDo: take care
    # here are notebooks which are not tested due to random issues
    blacklist = []

    for tutorial in tutorials:
        print(42 * "#", tutorial)
        # # for now we skip all tutorials but the one for which we have implemented SMOKE_TEST. This will change soon!
        if str(tutorial).split("/")[-1] in blacklist:
            print("Skipping", str(tutorial))
            continue
        num_runs += 1
        t1 = time.time()
        run_out = run_script(tutorial, env=env, timeout_minutes=timeout_minutes)
        elapsed_time = time.time() - t1
        print(f"time elapsed:{elapsed_time:.2f}")
        if run_out is None:  # in this case it bumped against max wall time
            df.loc[tutorial.name, "ran_successfully"] = False
            df.loc[tutorial.name, "message"] = "walltime exceeded"
            continue
        print(f"statuscode: {run_out.returncode}")

        if run_out.returncode != 0:
            num_errors += 1
            df.loc[tutorial.name, "message"] = run_out.stderr
            print(run_out.stderr)

        else:
            print(
                f"Running tutorial {tutorial.name} took " f"{elapsed_time:.2f} seconds."
            )
            df.loc[tutorial.name, "ran_successfully"] = True

    df.to_csv("notebook_test_stats.csv")

    # delete temporary test notebook file
    if os.path.exists("temp.ipynb"):
        os.remove("temp.ipynb")

    if num_errors > 0:
        raise RuntimeError(
            f"Running {num_runs} tutorials resulted in {num_errors} errors."
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Runs tutorials.")

    parser.add_argument(
        "-n",
        "--name",
        help="Run a specific tutorial by name.",
    )

    parser.add_argument(
        "-l",
        "--long",
        action="store_true",
        help="Run the full version of the notebook. Will take a long time",
    )

    # parser.add_argument(
    #     "-s", "--smoke", action="store_true", help="Run in smoke test (quick) mode."
    # )

    parser.add_argument(
        "-p", "--path", metavar="path", required=False, help="bofire repo directory."
    )

    args = parser.parse_args()

    run_tutorials(
        name=args.name,
        smoke_test=not args.long,
    )



import pytest
from pydantic.error_wrappers import ValidationError

import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from tests.bofire.strategies.test_qehvi import BOTORCH_QEHVI_STRATEGY_SPECS
from tests.bofire.strategies.test_qparego import BOTORCH_QPAREGO_STRATEGY_SPECS
from tests.bofire.strategies.test_sobo import (
    BOTORCH_ADDITIVE_AND_MULTIPLICATIVE_SOBO_STRATEGY_SPECS,
    BOTORCH_SOBO_STRATEGY_SPECS,
)

STRATEGY_SPECS = {
    data_models.SoboStrategy: BOTORCH_SOBO_STRATEGY_SPECS,
    data_models.AdditiveSoboStrategy: BOTORCH_ADDITIVE_AND_MULTIPLICATIVE_SOBO_STRATEGY_SPECS,
    data_models.MultiplicativeSoboStrategy: BOTORCH_ADDITIVE_AND_MULTIPLICATIVE_SOBO_STRATEGY_SPECS,
    data_models.QehviStrategy: BOTORCH_QEHVI_STRATEGY_SPECS,
    data_models.QnehviStrategy: BOTORCH_QEHVI_STRATEGY_SPECS,
    data_models.QparegoStrategy: BOTORCH_QPAREGO_STRATEGY_SPECS,
}


@pytest.mark.parametrize(
    "cls, spec",
    [(cls, valid) for cls, data in STRATEGY_SPECS.items() for valid in data["valids"]],
)
def test_valid_strategy_specs(cls, spec):
    res = cls(**spec)
    assert isinstance(res, cls)


@pytest.mark.parametrize(
    "cls, spec",
    [
        (cls, invalid)
        for cls, data in STRATEGY_SPECS.items()
        for invalid in data["invalids"]
    ],
)
def test_invalid_strategy_specs(cls, spec):
    with pytest.raises((ValueError, TypeError, KeyError, ValidationError)):
        print("mycls:", cls)
        print("invalid spec:", spec)
        data_model = cls(**spec)
        strategies.map(data_model=data_model)


# TODO: add per-strategy test for strategy.has_sufficient_experiments
# TODO: add per-strategy test for strategy.tell



import pytest

from bofire.data_models.api import AnyThing
from bofire.data_models.base import BaseModel


@pytest.mark.parametrize("model", AnyThing)
def test_domain_models_should_generate_schema(model: BaseModel):
    model.schema()



from typing import List, Literal, Union

from pydantic import Field, validator
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel


class Condition(BaseModel):
    type: str


class SingleCondition(BaseModel):
    type: str


class NumberOfExperimentsCondition(SingleCondition):
    type: Literal["NumberOfExperimentsCondition"] = "NumberOfExperimentsCondition"
    n_experiments: Annotated[int, Field(ge=1)]


class AlwaysTrueCondition(SingleCondition):
    type: Literal["AlwaysTrueCondition"] = "AlwaysTrueCondition"


class CombiCondition(Condition):
    type: Literal["CombiCondition"] = "CombiCondition"
    conditions: Annotated[
        List[
            Union[NumberOfExperimentsCondition, "CombiCondition", AlwaysTrueCondition]
        ],
        Field(min_items=2),
    ]
    n_required_conditions: Annotated[int, Field(ge=0)]

    @validator("n_required_conditions")
    def validate_n_required_conditions(cls, v, values):
        if v > len(values["conditions"]):
            raise ValueError(
                "Number of required conditions larger than number of conditions."
            )
        return v



from typing import List, Optional

import numpy as np
import torch
from botorch.acquisition.multi_objective.monte_carlo import (
    qExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.objective import (
    GenericMCMultiOutputObjective,
    MCMultiOutputObjective,
)
from botorch.utils.multi_objective.box_decompositions.non_dominated import (
    NondominatedPartitioning,
)

from bofire.data_models.objectives.api import ConstrainedObjective
from bofire.data_models.strategies.api import QehviStrategy as DataModel
from bofire.strategies.predictives.botorch import BotorchStrategy
from bofire.utils.multiobjective import get_ref_point_mask, infer_ref_point
from bofire.utils.torch_tools import get_multiobjective_objective


class QehviStrategy(BotorchStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.ref_point = data_model.ref_point
        self.ref_point_mask = get_ref_point_mask(self.domain)

    ref_point: Optional[dict] = None
    objective: Optional[MCMultiOutputObjective] = None

    def _get_acqfs(self, n) -> List[qExpectedHypervolumeImprovement]:
        df = self.domain.outputs.preprocess_experiments_all_valid_outputs(
            self.experiments
        )

        train_obj = (
            df[self.domain.outputs.get_keys_by_objective(excludes=None)].values
            * self.ref_point_mask
        )
        ref_point = self.get_adjusted_refpoint()
        weights = np.array(
            [
                feat.objective.w  # type: ignore
                for feat in self.domain.outputs.get_by_objective(excludes=None)
            ]
        )
        # compute points that are better than the known reference point
        better_than_ref = (train_obj > ref_point).all(axis=-1)
        # partition non-dominated space into disjoint rectangles
        partitioning = NondominatedPartitioning(
            ref_point=torch.from_numpy(ref_point * weights),
            # use observations that are better than the specified reference point and feasible
            Y=torch.from_numpy(train_obj[better_than_ref]),
        )

        _, X_pending = self.get_acqf_input_tensors()

        assert self.model is not None
        # setup the acqf
        acqf = qExpectedHypervolumeImprovement(
            model=self.model,
            ref_point=ref_point,  # use known reference point
            partitioning=partitioning,
            # sampler=self.sampler,
            # define an objective that specifies which outcomes are the objectives
            objective=self._get_objective(),
            X_pending=X_pending,
        )
        acqf._default_sample_shape = torch.Size([self.num_sobol_samples])
        return [acqf]

    def _get_objective(self) -> GenericMCMultiOutputObjective:
        objective = get_multiobjective_objective(outputs=self.domain.outputs)
        return GenericMCMultiOutputObjective(objective=objective)

    def get_adjusted_refpoint(self) -> List[float]:
        if self.ref_point is None:
            df = self.domain.outputs.preprocess_experiments_all_valid_outputs(
                self.experiments
            )
            ref_point = infer_ref_point(
                self.domain, experiments=df, return_masked=False
            )
        else:
            ref_point = self.ref_point
        return (
            self.ref_point_mask
            * np.array(
                [
                    ref_point[feat]
                    for feat in self.domain.outputs.get_keys_by_objective(
                        excludes=ConstrainedObjective
                    )
                ]
            )
        ).tolist()



from typing import List

import gpytorch
import gpytorch.kernels
import pytest
import torch
from pydantic import parse_obj_as

import bofire
import bofire.kernels.api as kernels
from bofire.data_models.kernels.api import (
    AdditiveKernel,
    AnyKernel,
    LinearKernel,
    MaternKernel,
    MultiplicativeKernel,
    PolynomialKernel,
    RBFKernel,
    ScaleKernel,
    TanimotoKernel,
)
from bofire.data_models.priors.api import BOTORCH_SCALE_PRIOR, GammaPrior


def get_invalids(valid: dict) -> List[dict]:
    return [
        {k: v for k, v in valid.items() if k != k_}
        for k_ in valid.keys()
        if k_ != "type"
    ]


EQUIVALENTS = {
    RBFKernel: gpytorch.kernels.RBFKernel,
    MaternKernel: gpytorch.kernels.MaternKernel,
    LinearKernel: gpytorch.kernels.LinearKernel,
    ScaleKernel: gpytorch.kernels.ScaleKernel,
    AdditiveKernel: gpytorch.kernels.AdditiveKernel,
    MultiplicativeKernel: gpytorch.kernels.ProductKernel,
    TanimotoKernel: bofire.kernels.fingerprint_kernels.tanimoto_kernel.TanimotoKernel,
}

VALID_RBF_SPEC = {
    "type": "RBFKernel",
    "ard": True,
}
VALID_MATERN_SPEC = {
    "type": "MaternKernel",
    "ard": True,
    "nu": 2.5,
}
VALID_LINEAR_SPEC = {"type": "LinearKernel"}

VALID_SCALE_SPEC = {"type": "ScaleKernel", "base_kernel": RBFKernel()}

VALID_ADDITIVE_SPEC = {"type": "AdditiveKernel", "kernels": [RBFKernel(), RBFKernel()]}

VALID_MULTIPLICATIVE_SPEC = {
    "type": "MultiplicativeKernel",
    "kernels": [RBFKernel(), RBFKernel()],
}

VALID_TANIMOTO_SPEC = {
    "type": "TanimotoKernel",
    "ard": True,
}

KERNEL_SPECS = {
    RBFKernel: {
        "valids": [
            VALID_RBF_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_RBF_SPEC),
        ],
    },
    MaternKernel: {
        "valids": [
            VALID_MATERN_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_MATERN_SPEC),
        ],
    },
    LinearKernel: {
        "valids": [
            VALID_LINEAR_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_LINEAR_SPEC),
        ],
    },
    ScaleKernel: {
        "valids": [VALID_SCALE_SPEC],
        "invalids": [
            *get_invalids(VALID_SCALE_SPEC),
        ],
    },
    MultiplicativeKernel: {
        "valids": [VALID_MULTIPLICATIVE_SPEC],
        "invalids": [
            *get_invalids(VALID_SCALE_SPEC),
        ],
    },
    AdditiveKernel: {
        "valids": [VALID_ADDITIVE_SPEC],
        "invalids": [
            *get_invalids(VALID_SCALE_SPEC),
        ],
    },
    TanimotoKernel: {
        "valids": [VALID_TANIMOTO_SPEC],
        "invalids": [
            *get_invalids(VALID_TANIMOTO_SPEC),
        ],
    },
}


@pytest.mark.parametrize(
    "cls, spec",
    [(cls, valid) for cls, data in KERNEL_SPECS.items() for valid in data["valids"]],
)
def test_valid_kernel_specs(cls, spec):
    res = cls(**spec)
    assert isinstance(res, cls)
    assert isinstance(res.__str__(), str)
    gkernel = kernels.map(
        res, batch_shape=torch.Size(), ard_num_dims=10, active_dims=list(range(5))
    )
    assert isinstance(gkernel, EQUIVALENTS[cls])
    res2 = parse_obj_as(AnyKernel, res.dict())
    assert res == res2


def test_scale_kernel():
    kernel = ScaleKernel(
        base_kernel=RBFKernel(), outputscale_prior=BOTORCH_SCALE_PRIOR()
    )
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=10,
        active_dims=list(range(5)),
    )
    assert hasattr(k, "outputscale_prior")
    assert isinstance(k.outputscale_prior, gpytorch.priors.GammaPrior)
    kernel = ScaleKernel(base_kernel=RBFKernel())
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=10,
        active_dims=list(range(5)),
    )
    assert hasattr(k, "outputscale_prior") is False


def test_poly_kernel():
    kernel = PolynomialKernel(power=2, offset_prior=BOTORCH_SCALE_PRIOR())
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=10,
        active_dims=list(range(5)),
    )
    assert hasattr(k, "offset_prior")
    assert isinstance(k.offset_prior, gpytorch.priors.GammaPrior)
    kernel = PolynomialKernel(power=2)
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=10,
        active_dims=list(range(5)),
    )
    assert hasattr(k, "offset_prior") is False


@pytest.mark.parametrize(
    "kernel, ard_num_dims, active_dims, expected_kernel",
    [
        (
            RBFKernel(
                ard=False, lengthscale_prior=GammaPrior(concentration=2.0, rate=0.15)
            ),
            10,
            list(range(5)),
            gpytorch.kernels.RBFKernel,
        ),
        (
            RBFKernel(ard=False),
            10,
            list(range(5)),
            gpytorch.kernels.RBFKernel,
        ),
        (RBFKernel(ard=True), 10, list(range(5)), gpytorch.kernels.RBFKernel),
        (
            MaternKernel(
                ard=False, lengthscale_prior=GammaPrior(concentration=2.0, rate=0.15)
            ),
            10,
            list(range(5)),
            gpytorch.kernels.MaternKernel,
        ),
        (MaternKernel(ard=False), 10, list(range(5)), gpytorch.kernels.MaternKernel),
        (MaternKernel(ard=True), 10, list(range(5)), gpytorch.kernels.MaternKernel),
        (
            MaternKernel(ard=False, nu=2.5),
            10,
            list(range(5)),
            gpytorch.kernels.MaternKernel,
        ),
        (
            MaternKernel(ard=True, nu=1.5),
            10,
            list(range(5)),
            gpytorch.kernels.MaternKernel,
        ),
        (LinearKernel(), 10, list(range(5)), gpytorch.kernels.LinearKernel),
    ],
)
def test_continuous_kernel(kernel, ard_num_dims, active_dims, expected_kernel):
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=ard_num_dims,
        active_dims=active_dims,
    )
    assert isinstance(k, expected_kernel)
    if isinstance(kernel, LinearKernel):
        return
    if kernel.lengthscale_prior is not None:
        assert hasattr(k, "lengthscale_prior")
        assert isinstance(k.lengthscale_prior, gpytorch.priors.GammaPrior)
    else:
        assert hasattr(k, "lengthscale_prior") is False

    if kernel.ard is False:
        assert k.ard_num_dims is None
    else:
        assert k.ard_num_dims == len(active_dims)
    assert torch.eq(k.active_dims, torch.tensor(active_dims, dtype=torch.int64)).all()

    if isinstance(kernel, gpytorch.kernels.MaternKernel):
        assert kernel.nu == k.nu


@pytest.mark.parametrize(
    "kernel, ard_num_dims, active_dims, expected_kernel",
    [
        (
            TanimotoKernel(ard=False),
            10,
            list(range(5)),
            bofire.kernels.fingerprint_kernels.tanimoto_kernel.TanimotoKernel,
        ),
        (
            TanimotoKernel(ard=True),
            10,
            list(range(5)),
            bofire.kernels.fingerprint_kernels.tanimoto_kernel.TanimotoKernel,
        ),
    ],
)
def test_molecular_kernel(kernel, ard_num_dims, active_dims, expected_kernel):
    k = kernels.map(
        kernel,
        batch_shape=torch.Size(),
        ard_num_dims=ard_num_dims,
        active_dims=active_dims,
    )
    assert isinstance(k, expected_kernel)

    if kernel.ard is False:
        assert k.ard_num_dims is None
    else:
        assert k.ard_num_dims == len(active_dims)
    assert torch.eq(k.active_dims, torch.tensor(active_dims, dtype=torch.int64)).all()



import os.path

from setuptools import find_packages, setup

sklearn_dependency = "scikit-learn>=1.0.0"

root_dir = os.path.dirname(__file__)
with open(os.path.join(root_dir, "README.md"), "r") as f:
    long_description = f.read()


setup(
    name="bofire",
    description="",
    author="",
    license="BSD-3",
    url="https://github.com/experimental-design/bofire",
    keywords=[
        "Bayesian optimization",
        "Multi-objective optimization",
        "Experimental design",
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.9.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy",
        "pandas",
        "pydantic>=1.10.0,<2.0",
        "scipy>=1.7",
        "typing-extensions",
    ],
    extras_require={
        "optimization": [
            "botorch>=0.9.4",
            "multiprocess",
            "plotly",
            "formulaic>=0.6.0",
            "cloudpickle>=2.0.0",
            "sympy>=1.12",
            sklearn_dependency,
        ],
        "xgb": ["xgboost>=1.7.5"],
        "cheminfo": ["rdkit>=2023.3.2", sklearn_dependency, "mordred"],
        "tests": [
            "mock",
            "mopti",
            "pyright==1.1.305",
            "pytest",
            "pytest-cov",
            "papermill",
            "jupyter",
            "matplotlib",
        ],
        "docs": [
            "mkdocs",
            "mkdocs-material",
            "mkdocs-jupyter",
            "mkdocstrings>=0.18",
            "mkdocstrings-python-legacy",
            "mike",
        ],
    },
)



from typing import Optional

import pandas as pd

from bofire.data_models.strategies.api import FactorialStrategy as DataModel
from bofire.strategies.strategy import Strategy


class FactorialStrategy(Strategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _ask(self, candidate_count: Optional[int] = None) -> pd.DataFrame:
        if candidate_count is not None:
            raise ValueError(
                "FactorialStrategy will ignore the specified value of candidate_count. "
                "The strategy automatically determines how many candidates to "
                "propose."
            )
        return pd.DataFrame.from_dict(
            [
                {e[0]: e[1] for e in combi}  # type: ignore
                for combi in self.domain.inputs.get_categorical_combinations()
            ]
        )

    def has_sufficient_experiments(self) -> bool:
        return True



import random

import bofire.data_models.acquisition_functions.api as acquisition_functions
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    acquisition_functions.qEI,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qLogEI,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qNEI,
    lambda: {"prune_baseline": random.choice([True, False])},
)

specs.add_valid(
    acquisition_functions.qLogNEI,
    lambda: {"prune_baseline": random.choice([True, False])},
)


specs.add_valid(
    acquisition_functions.qPI,
    lambda: {
        "tau": random.random(),
    },
)

specs.add_valid(
    acquisition_functions.qSR,
    lambda: {},
)

specs.add_valid(
    acquisition_functions.qUCB,
    lambda: {
        "beta": random.random(),
    },
)

specs.add_valid(
    acquisition_functions.qEHVI,
    lambda: {
        "alpha": random.random(),
    },
)

specs.add_valid(
    acquisition_functions.qLogEHVI,
    lambda: {
        "alpha": random.random(),
    },
)

specs.add_valid(
    acquisition_functions.qNEHVI,
    lambda: {"alpha": random.random(), "prune_baseline": random.choice([True, False])},
)

specs.add_valid(
    acquisition_functions.qLogNEHVI,
    lambda: {"alpha": random.random(), "prune_baseline": random.choice([True, False])},
)



import os
import re
import sys
from tempfile import gettempdir
from uuid import uuid4


def test_docs():
    """This tests tests the codeblocks embedded within the documentation"""
    folder_under_test = "docs"
    files_list = [
        os.path.join(folder_under_test, x)
        for x in os.listdir(folder_under_test)
        if x.rsplit(".")[-1] == "md"
    ]
    files_under_test = files_list + ["README.md"]
    code_block_start = "```python"
    code_block_end = "```"
    code_block_regex = re.compile(
        f"{code_block_start}(.*?){code_block_end}", flags=re.DOTALL
    )
    for filename in files_under_test:
        print(f"### Processing doc file {filename} ###")
        with open(filename, "r") as f:
            content = f.read()
        codeblocks = code_block_regex.findall(content)
        codeblocks = "\n".join(codeblocks)
        tmpfile = f"{gettempdir()}/{uuid4().hex}.py"
        exit_code = None
        try:
            with open(tmpfile, "w") as f:
                f.write(codeblocks)
            exit_code = os.system(f"{sys.executable} {tmpfile}")
        finally:
            os.remove(tmpfile)
        assert exit_code == 0



from bofire.data_models.strategies.strategy import Strategy


class SamplerStrategy(Strategy):
    """Base class for sampling methods in BoFire for sampling from constrained input spaces."""

    pass



from typing import Literal, Type

from bofire.data_models.constraints.api import (
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.enum import SamplingMethodEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    CategoricalMolecularInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
)
from bofire.data_models.strategies.samplers.sampler import SamplerStrategy


class RejectionSampler(SamplerStrategy):
    """Sampler that generates samples via rejection sampling.

    Attributes:
        domain (Domain): Domain defining the constrained input space
        sampling_method (SamplingMethodEnum, optional): Method to generate the base samples. Defaults to UNIFORM.
        num_base_samples (int, optional): Number of base samples to sample in each iteration. Defaults to 1000.
        max_iters (int, optinal): Number of iterations. Defaults to 1000.
    """

    type: Literal["RejectionSampler"] = "RejectionSampler"
    sampling_method: SamplingMethodEnum = SamplingMethodEnum.UNIFORM
    num_base_samples: int = 1000
    max_iters: int = 1000

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            LinearInequalityConstraint,
            NonlinearInequalityConstraint,
            NChooseKConstraint,
        ]

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            ContinuousInput,
            ContinuousOutput,
            DiscreteInput,
            CategoricalInput,
            CategoricalDescriptorInput,
            CategoricalMolecularInput,
        ]



from typing import Literal

from pydantic import PositiveFloat

from bofire.data_models.priors.prior import Prior


class GammaPrior(Prior):
    """Gamma prior based on the gamma distribution

    Attributes:
        concentration(PostiveFloat): concentration of the gamma distribution
        rate(PositiveFloat): rate of the gamma prior.
    """

    type: Literal["GammaPrior"] = "GammaPrior"
    concentration: PositiveFloat
    rate: PositiveFloat



from typing import Callable, List, Tuple, Union

import torch
from botorch.acquisition import get_acquisition_function
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.objective import GenericMCObjective
from botorch.models.gpytorch import GPyTorchModel
from botorch.utils.multi_objective.scalarization import get_chebyshev_scalarization
from botorch.utils.sampling import sample_simplex

from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MinimizeObjective,
    Objective,
)
from bofire.data_models.strategies.api import QparegoStrategy as DataModel
from bofire.strategies.predictives.botorch import BotorchStrategy
from bofire.utils.multiobjective import get_ref_point_mask
from bofire.utils.torch_tools import (
    get_multiobjective_objective,
    get_output_constraints,
    tkwargs,
)


# this implementation follows this tutorial: https://github.com/pytorch/botorch/blob/main/tutorials/multi_objective_bo.ipynb
# main difference to the multiobjective strategies is that we have a randomized list of acqfs, this has to be bring into accordance
# with the other strategies
class QparegoStrategy(BotorchStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.acquisition_function = data_model.acquisition_function
        self.constraint_callables, self.etas = None, 1e-3

    def _get_objective_and_constraints(
        self,
    ) -> Tuple[
        GenericMCObjective,
        Union[List[Callable[[torch.Tensor], torch.Tensor]], None],
        Union[List, float],
    ]:
        """Returns the scalarized objective.

        Returns:
            GenericMCObjective: the botorch objective.
            Union[ConstrainedObjective, None]: the botorch constraints.
            Union[List, float]: etas used in the botorch constraints.
        """
        ref_point_mask = torch.from_numpy(get_ref_point_mask(domain=self.domain)).to(
            **tkwargs
        )
        weights = (
            sample_simplex(
                len(
                    self.domain.outputs.get_keys_by_objective(
                        includes=[
                            MaximizeObjective,
                            MinimizeObjective,
                            CloseToTargetObjective,
                        ]
                    )
                ),
                **tkwargs,
            ).squeeze()
            * ref_point_mask
        )

        obj_callable = get_multiobjective_objective(outputs=self.domain.outputs)

        df_preds = self.predict(
            self.domain.outputs.preprocess_experiments_any_valid_output(
                experiments=self.experiments
            )
        )

        preds = torch.from_numpy(
            df_preds[[f"{key}_pred" for key in self.domain.outputs.get_keys()]].values
        ).to(**tkwargs)

        scalarization = get_chebyshev_scalarization(
            weights=weights, Y=obj_callable(preds, None) * ref_point_mask
        )

        def objective_callable(Z, X=None):
            return scalarization(obj_callable(Z, None) * ref_point_mask, X)

        if len(weights) != len(self.domain.outputs.get_by_objective(Objective)):
            constraint_callables, etas = get_output_constraints(self.domain.outputs)
        else:
            constraint_callables, etas = None, 1e-3

        return (
            GenericMCObjective(objective=objective_callable),
            constraint_callables,
            etas,
        )

    def _get_acqfs(self, n: int) -> List[AcquisitionFunction]:
        assert self.is_fitted is True, "Model not trained."

        acqfs = []

        X_train, X_pending = self.get_acqf_input_tensors()

        (
            objective_callable,
            constraint_callables,
            etas,
        ) = self._get_objective_and_constraints()

        assert self.model is not None
        for i in range(n):
            acqf = get_acquisition_function(
                acquisition_function_name=self.acquisition_function.__class__.__name__,
                model=self.model,
                objective=objective_callable,
                X_observed=X_train,
                X_pending=X_pending if i == 0 else None,
                constraints=constraint_callables,
                eta=torch.tensor(etas).to(**tkwargs),
                mc_samples=self.num_sobol_samples,
                prune_baseline=True,
                cache_root=True if isinstance(self.model, GPyTorchModel) else False,
            )
            acqfs.append(acqf)
        return acqfs



from tests.bofire.data_models.specs.api import Spec


def test_prior_should_be_serializable(prior_spec: Spec):
    spec = prior_spec.typed_spec()
    obj = prior_spec.cls(**spec)
    assert obj.dict() == spec


def test_kernel_should_be_serializable(kernel_spec: Spec):
    spec = kernel_spec.typed_spec()
    obj = kernel_spec.cls(**spec)
    assert obj.dict() == spec


def test_constraint_should_be_serializable(constraint_spec: Spec):
    spec = constraint_spec.typed_spec()
    obj = constraint_spec.cls(**spec)
    assert obj.dict() == spec


def test_objective_should_be_serializable(objective_spec: Spec):
    spec = objective_spec.typed_spec()
    obj = objective_spec.cls(**spec)
    assert obj.dict() == spec


def test_feature_should_be_serializable(feature_spec: Spec):
    spec = feature_spec.typed_spec()
    obj = feature_spec.cls(**spec)
    print(spec)
    assert obj.dict() == spec


def test_domain_should_be_serializable(domain_spec: Spec):
    spec = domain_spec.typed_spec()
    obj = domain_spec.cls(**spec)
    assert obj.dict() == spec


def test_surrogate_should_be_serializable(surrogate_spec: Spec):
    spec = surrogate_spec.typed_spec()
    obj = surrogate_spec.cls(**spec)
    assert obj.dict() == spec


def test_acquisition_function_should_be_serializable(acquisition_function_spec: Spec):
    spec = acquisition_function_spec.typed_spec()
    obj = acquisition_function_spec.cls(**spec)
    assert obj.dict() == spec


def test_strategy_should_be_serializable(strategy_spec: Spec):
    spec = strategy_spec.typed_spec()
    obj = strategy_spec.cls(**spec)
    # TODO: can we unhide the comparison of surrogate_specs?
    data = {k: v for k, v in obj.dict().items() if k != "surrogate_specs"}
    for k, v in data.items():
        if v is not None:
            assert v == spec[k]


def test_condition_should_be_serializable(condition_spec: Spec):
    spec = condition_spec.typed_spec()
    obj = condition_spec.cls(**spec)
    assert obj.dict() == spec


def test_outlier_detection_should_be_serializable(outlier_detection_spec: Spec):
    spec = outlier_detection_spec.typed_spec()
    obj = outlier_detection_spec.cls(**spec)
    assert obj.dict() == spec


def test_molfeatures_should_be_serializable(molfeatures_spec: Spec):
    spec = molfeatures_spec.typed_spec()
    obj = molfeatures_spec.cls(**spec)
    assert obj.dict() == spec



from typing import ClassVar, Literal

import numpy as np
import pandas as pd
from pydantic import validator

from bofire.data_models.features.feature import TDiscreteVals
from bofire.data_models.features.numerical import NumericalInput


class DiscreteInput(NumericalInput):
    """Feature with discretized ordinal values allowed in the optimization.

    Attributes:
        key(str): key of the feature.
        values(List[float]): the discretized allowed values during the optimization.
    """

    type: Literal["DiscreteInput"] = "DiscreteInput"
    order_id: ClassVar[int] = 3

    values: TDiscreteVals

    @validator("values")
    def validate_values_unique(cls, values):
        """Validates that provided values are unique.

        Args:
            values (List[float]): List of values

        Raises:
            ValueError: when values are non-unique.
            ValueError: when values contains only one entry.
            ValueError: when values is empty.

        Returns:
            List[values]: Sorted list of values
        """
        if len(values) != len(set(values)):
            raise ValueError("Discrete values must be unique")
        if len(values) == 1:
            raise ValueError(
                "Fixed discrete inputs are not supported. Please use a fixed continuous input."
            )
        if len(values) == 0:
            raise ValueError("No values defined.")
        return sorted(values)

    @property
    def lower_bound(self) -> float:
        """Lower bound of the set of allowed values"""
        return min(self.values)

    @property
    def upper_bound(self) -> float:
        """Upper bound of the set of allowed values"""
        return max(self.values)

    def validate_candidental(self, values: pd.Series) -> pd.Series:
        """Method to validate the provided candidates.

        Args:
            values (pd.Series): suggested candidates for the feature

        Raises:
            ValueError: Raises error when one of the provided values is not contained in the list of allowed values.

        Returns:
            pd.Series: _uggested candidates for the feature
        """
        values = super().validate_candidental(values)
        if not np.isin(values.to_numpy(), np.array(self.values)).all():
            raise ValueError(
                f"Not allowed values in candidates for feature {self.key}."
            )
        return values

    def sample(self, n: int) -> pd.Series:
        """Draw random samples from the feature.

        Args:
            n (int): number of samples.

        Returns:
            pd.Series: drawn samples.
        """
        return pd.Series(name=self.key, data=np.random.choice(self.values, n))

    def from_continuous(self, values: pd.DataFrame) -> pd.Series:
        """Rounds continuous values to the closest discrete ones.

        Args:
            values (pd.DataFrame): Dataframe with continuous entries.

        Returns:
            pd.Series: Series with discrete values.
        """

        s = pd.DataFrame(
            data=np.abs(
                (values[self.key].to_numpy()[:, np.newaxis] - np.array(self.values))
            ),
            columns=self.values,
            index=values.index,
        ).idxmin(1)
        s.name = self.key
        return s



from typing import Literal

from pydantic import PositiveFloat

from bofire.data_models.priors.prior import Prior


class NormalPrior(Prior):
    """Normal prior based on the normal distribution

    Attributes:
        loc(float): mean/center of the normal distribution
        scale(PositiveFloat): width of the normal distribution
    """

    type: Literal["NormalPrior"] = "NormalPrior"
    loc: float
    scale: PositiveFloat



import random
from itertools import chain

import numpy as np
import pytest
import torch
from botorch.acquisition.multi_objective import (
    qExpectedHypervolumeImprovement,
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.objective import GenericMCMultiOutputObjective

import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from bofire.benchmarks.multi import C2DTLZ2, DTLZ2
from bofire.data_models.domain.api import Domain
from bofire.data_models.enum import CategoricalMethodEnum
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.objectives.api import MaximizeObjective, MinimizeObjective
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.api import PolytopeSampler
from tests.bofire.strategies.specs import VALID_CONTINUOUS_INPUT_FEATURE_SPEC
from tests.bofire.utils.test_multiobjective import (
    dfs,
    invalid_domains,
    valid_constrained_domains,
    valid_domains,
)

if1 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if1",
    }
)

if2 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if2",
    }
)

if3 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if3",
    }
)

of1 = ContinuousOutput(
    objective=MaximizeObjective(w=1),
    key="of1",
)

of2 = ContinuousOutput(
    objective=MinimizeObjective(w=1),
    key="of2",
)

VALID_BOTORCH_QEHVI_STRATEGY_SPEC = {
    "domain": Domain.from_lists(inputs=[if1, if2, if3], outputs=[of1, of2]),
    # "num_sobol_samples": 1024,
    # "num_restarts": 8,
    # "num_raw_samples": 1024,
    "descriptor_method": random.choice(list(CategoricalMethodEnum)),
    "categorical_method": "EXHAUSTIVE",
}

BOTORCH_QEHVI_STRATEGY_SPECS = {
    "valids": [
        VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
        {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "seed": 1},
        # {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "surrogate_specs": VALID_MODEL_SPEC_LIST},
        {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "ref_point": {"of1": 1.0, "of2": 2}},
    ],
    "invalids": [
        {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "descriptor_method": None},
        {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "categorical_method": None},
        # {**VALID_BOTORCH_QEHVI_STRATEGY_SPEC, "seed": -1},
    ],
}


@pytest.mark.parametrize(
    "domain, ref_point",
    [
        (invalid_domains[0], None),
        (invalid_domains[1], None),
        (valid_domains[0], [0]),
        (valid_domains[0], {}),
        (valid_domains[0], {"of1": 0.0, "of2": 0, "of3": 0}),
        (valid_domains[0], {"of1": 0.0}),
        (valid_domains[0], {"of1": 0.0, "of3": 0.0}),
    ],
)
def test_invalid_qehvi(domain, ref_point):
    with pytest.raises(ValueError):
        data_models.QehviStrategy(domain=domain, ref_point=ref_point)


@pytest.mark.parametrize("domain", valid_constrained_domains)
def test_qehvi_invalid_constrained_objectives(domain):
    with pytest.raises(ValueError):
        data_models.QehviStrategy(domain=domain)


@pytest.mark.parametrize("domain", valid_constrained_domains)
def test_qnehvi_valid_constrained_objectives(domain):
    data_models.QnehviStrategy(domain=domain)


@pytest.mark.parametrize(
    "domain, ref_point, experiments, expected",
    [
        (valid_domains[0], {"of1": 0.5, "of2": 10.0}, dfs[0], [0.5, -10.0]),
        (valid_domains[1], {"of1": 0.5, "of3": 0.5}, dfs[1], [0.5, 0.5]),
        (valid_domains[0], None, dfs[0], [1.0, -5.0]),
        (valid_domains[1], None, dfs[1], [1.0, 2.0]),
    ],
)
def test_qehvi_get_adjusted_refpoint(domain, ref_point, experiments, expected):
    data_model = data_models.QehviStrategy(domain=domain, ref_point=ref_point)
    strategy = strategies.map(data_model)
    # hack for the test to prevent training of the model when using tell
    strategy.set_experiments(experiments)
    adjusted_ref_point = strategy.get_adjusted_refpoint()
    assert isinstance(adjusted_ref_point, list)
    assert np.allclose(expected, np.asarray(adjusted_ref_point))


@pytest.mark.parametrize(
    "strategy, use_ref_point, num_test_candidates",
    [
        (strategy, use_ref_point, num_test_candidates)
        for strategy in [data_models.QehviStrategy, data_models.QnehviStrategy]
        for use_ref_point in [True, False]
        for num_test_candidates in range(1, 3)
    ],
)
def test_qehvi(strategy, use_ref_point, num_test_candidates):
    # generate data
    benchmark = DTLZ2(dim=6)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    # init strategy
    data_model = strategy(
        domain=benchmark.domain,
        ref_point=benchmark.ref_point if use_ref_point else None,
    )
    my_strategy = strategies.map(data_model)
    my_strategy.tell(experiments)

    acqf = my_strategy._get_acqfs(2)[0]

    assert isinstance(acqf.objective, GenericMCMultiOutputObjective)
    assert isinstance(
        acqf,
        qExpectedHypervolumeImprovement
        if strategy == data_models.QehviStrategy
        else qNoisyExpectedHypervolumeImprovement,
    )
    # test acqf calc
    # acqf_vals = my_strategy._choose_from_pool(experiments_test, num_test_candidates)
    # assert acqf_vals.shape[0] == num_test_candidates


def test_qnehvi_constraints():
    benchmark = C2DTLZ2(dim=4)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    data_model = data_models.QnehviStrategy(
        domain=benchmark.domain, ref_point={"f_0": 1.1, "f_1": 1.1}
    )
    my_strategy = strategies.map(data_model)
    my_strategy.tell(experiments)
    acqf = my_strategy._get_acqfs(2)[0]
    assert isinstance(acqf.objective, GenericMCMultiOutputObjective)
    assert isinstance(acqf, qNoisyExpectedHypervolumeImprovement)
    assert acqf.eta == torch.tensor(1e-3)
    assert len(acqf.constraints) == 1
    assert torch.allclose(
        acqf.ref_point,
        torch.tensor([-1.1, -1.1], dtype=torch.double),
    )


@pytest.mark.parametrize(
    "strategy, ref_point, num_experiments, num_candidates",
    [
        (strategy, ref_point, num_experiments, num_candidates)
        for strategy in [data_models.QehviStrategy, data_models.QnehviStrategy]
        for ref_point in [None, {"f_0": 1.1, "f_1": 1.1}]
        for num_experiments in range(8, 10)
        for num_candidates in range(1, 3)
    ],
)
@pytest.mark.slow
def test_get_acqf_input(strategy, ref_point, num_experiments, num_candidates):
    # generate data
    benchmark = DTLZ2(dim=6)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(
        random_strategy._ask(n=num_experiments), return_complete=True
    )
    data_model = strategy(domain=benchmark.domain)
    strategy = strategies.map(data_model)
    # , ref_point=ref_pointw

    strategy.tell(experiments)
    strategy.ask(candidate_count=num_candidates, add_pending=True)

    X_train, X_pending = strategy.get_acqf_input_tensors()

    _, names = strategy.domain.inputs._get_transform_info(
        specs=strategy.surrogate_specs.input_preprocessing_specs
    )

    assert torch.is_tensor(X_train)
    assert torch.is_tensor(X_pending)
    assert X_train.shape == (
        num_experiments,
        len(set(chain(*names.values()))),
    )
    assert X_pending.shape == (
        num_candidates,
        len(set(chain(*names.values()))),
    )


def test_no_objective():
    domain = DTLZ2(dim=6).domain
    experiments = DTLZ2(dim=6).f(domain.inputs.sample(10), return_complete=True)
    domain.outputs.features.append(ContinuousOutput(key="ignore", objective=None))
    experiments["ignore"] = experiments["f_0"] + 6
    experiments["valid_ignore"] = 1
    data_model = data_models.QehviStrategy(
        domain=domain, ref_point={"f_0": 1.1, "f_1": 1.1}
    )
    recommender = strategies.map(data_model=data_model)
    recommender.tell(experiments=experiments)
    candidates = recommender.ask(candidate_count=1)
    recommender.to_candidates(candidates)
